from PIL import Image
import numpy as np

def get_end_directions(info_dict, pixel_coordinates, geodesics):
    """Internal use"""

    height, width, sample = info_dict['height'], info_dict['width'], info_dict['samples']
    
    k_x_end = np.zeros(width * height * 1)
    k_x_end.shape = height, width
    k_y_end = np.zeros(width * height  * 1)
    k_y_end.shape = height, width
    k_z_end = np.zeros(width * height * 1)
    k_z_end.shape = height, width

    for i in range(len(pixel_coordinates)):
        y, x, s = pixel_coordinates[i]
        if s == 0:
            k_xyz, x_xyz = geodesics[i]
            k_xyz = np.column_stack(k_xyz)
            k_x_end[y, x] = k_xyz[-1][0]
            k_y_end[y, x] = k_xyz[-1][1]
            k_z_end[y, x] = k_xyz[-1][2]

    return k_x_end, k_y_end, k_z_end

def projection(background_image, camera):
    """(Experimental) Calculates the projection of a skydome to the camera and returns a PIL Image"""

    k_x_end, k_y_end, k_z_end = get_end_directions(camera.cam_information_dict(), camera.pixel_coordinates, camera.geodesics)

    return simple_projection(background_image, k_x_end, k_y_end, k_z_end, camera.ray_blackhole_hit )


def simple_projection(background_image, k_x, k_y, k_z, hit_blackhole):
    """Internal use"""
    
    resy = k_x.shape[0]
    resx = k_x.shape[1]

    pil_im = Image.open(background_image)
    width, height = pil_im.size
    pixels = np.array(pil_im)#.getdata())
    #pixels = pixels.reshape(width,height,4)
    channels = pixels.shape[2]

    if channels == 3: 
        black = np.array([ 0,0,0], dtype=np.uint8)
    else:
        black = np.array([ 0,0,0,255], dtype=np.uint8)

    def background_hit(direction):
        color = np.zeros(3)  
        if direction[2] > 1 or direction[2]<-1:
            direction = direction / np.linalg.norm(direction)
    
        #arccos loopt van [0, pi]
        theta = np.arccos(direction[2])/np.pi # [0,1]
        #arctan2 loopt van [-pi, pi]
        phi = 1/2 - (1/2*np.arctan2(direction[1], direction[0])/np.pi) # [0,1]

        return pixels[int(theta*(height-1)), int((1-phi)*(width-1))]

    projection = np.zeros((resy, resx, channels))
    
    for iy, ix in np.ndindex(k_x.shape):
        if hit_blackhole[iy,ix,0] == 1.0:
            projection[iy, ix] = black
        else:
            projection[iy, ix] = background_hit([k_x[iy, ix], k_y[iy, ix], k_z[iy, ix]] )
    
    return Image.fromarray(projection.astype('uint8'))






import sympy as sp
import numpy as np
from scipy.integrate import solve_ivp
import time
from curvedpy.metrics.schwarzschild_metric import SchwarzschildMetricXYZ

from curvedpy.utils.conversions import Conversions

# -------------------------------
# ----NAMING CONVENTIONS USED----
# -------------------------------
#
# UPPER & LOWER INDICES
# _ indicates an upper/contravariant index
# __ indicates a lower/covariant index
# For example g__mu__nu is the metric g with two lower indices while g_mu_nu is the metric 
# with two upper indices, which is the inverse of g__mu__nu
# Another example is 4-momentum k_mu, which has a 0 (time) component of k_0 or k_t
# The 4-momentum as a oneform is then k__mu and its zeroth component is k__0
# An example tensor could be T_mu__nu__alp. This tensor has one upper/contravariant (mu) and 
# two lower/covariant/oneform (nu, alp) indices.
#
# COORDINATES
# the used coordinate system for a tensor is indicated with by appending for example 
# _xyz, _sph, _bl after the indices.
# Example: x_mu_bl: this 4vector with one upper index is given in Boyer-Lindquist coordinates
# Example: g__mu__nu_sph: this metric tensor with two lower indices is given in spherical coordinates
#
# MISC
# If a vector x_mu has only 3 components, they are the three spatial components


# https://f.yukterez.net/einstein.equations/files/8.html#transformation
# https://physics.stackexchange.com/questions/672252/how-to-compute-and-interpret-schwartzchild-black-hole-metric-line-element-ds-i
class GeodesicIntegratorSchwarzschildXYZ:

    conversions = Conversions()

    ################################################################################################
    #
    ################################################################################################
    def __init__(self, mass=1.0, time_like = False, auto_grad=True, verbose=False):

        self.M = mass
        self.r_s_value = 2*self.M 

        # Type of geodesic
        self.time_like = time_like # No Massive particle geodesics yet
        self.verbose = verbose
        if verbose:
            print("Geodesic SS Integrator Settings: ")
            print(f"  - {self.M=}")
            print(f"  - {self.r_s_value=}")
            print(f"  - {self.time_like=}")
            print(f"  - {self.verbose=}")
            print("--")

        self.metric = SchwarzschildMetricXYZ(mass, time_like, verbose)
        
        # # Define symbolic variables
        # #self.t, self.r, self.th, self.ph, self.r_s = sp.symbols("t r \\theta \\phi r_s", real=True)
        # self.t, self.x, self.y, self.z, = sp.symbols('t x y z', real=True)
        # self.r, self.r_s  = sp.symbols('r r_s', positive=True, real=True)
        # self.alp = sp.symbols('alp')
        

        # self.r_sub = (self.x**2 + self.y**2 + self.z**2)**sp.Rational(1,2)
        # self.alp_sub = self.r_s/(self.r**2*(-self.r_s+self.r)) # NOTE THE r_s is in here for most entries!

        # flip_sign_convention = -1

        # g00 = -1*( 1-self.r_s/self.r )
        # g01, g02, g03 = 0,0,0
        # g10 = 0
        # g11 = -1*( -1-self.x**2 * self.alp )
        # g12 = flip_sign_convention*( -self.x*self.y*self.alp )
        # g13 = flip_sign_convention*( -self.x*self.z*self.alp )
        # g20 = 0
        # g21 = flip_sign_convention*( -self.x*self.y*self.alp )
        # g22 = -1*( -1-self.y**2*self.alp )
        # g23 = flip_sign_convention*( -self.y*self.z*self.alp )
        # g30 = 0
        # g31 = flip_sign_convention*( -self.x*self.z*self.alp )
        # g32 = flip_sign_convention*( -self.y*self.z*self.alp )
        # g33 = -1*( -1 - self.z**2*self.alp )

        # self.g__mu__nu_cart = sp.Matrix([[g00,g01,g02,g03], [g10,g11,g12,g13], [g20,g21,g22,g23], [g30,g31,g32,g33]])
        # self.g__mu__nu_cart_pre_sub = self.g__mu__nu_cart

        # self.g__mu__nu_cart = self.g__mu__nu_cart.subs(self.alp, self.alp_sub).subs(self.r, self.r_sub)

        # g_00 = -1*( 1/(1-self.r_s/self.r) )
        # g_01, g_02, g_03 = 0,0,0
        # g_10 = 0
        # g_11 = -1*(-1+self.r_s  * self.x**2/self.r**3)
        # g_12 = flip_sign_convention*self.r_s * self.x * self.y/self.r**3
        # g_13 = flip_sign_convention*self.r_s * self.x * self.z/self.r**3
        # g_20 = 0
        # g_21 = flip_sign_convention* self.r_s * self.x * self.y/self.r**3
        # g_22 = -1*( -1 + self.r_s * self.y**2/self.r**3 )
        # g_23 = flip_sign_convention*( self.r_s * self.y * self.z/self.r**3 )
        # g_30 = 0
        # g_31 = flip_sign_convention* ( self.r_s * self.x * self.z/self.r**3 )
        # g_32 = flip_sign_convention*( self.r_s * self.y * self.z/self.r**3 )
        # g_33 = -1*( -1 + self.r_s * self.z**2/self.r**3 )

        # self.g_mu_nu_cart = sp.Matrix([[g_00,g_01,g_02,g_03], [g_10,g_11,g_12,g_13], [g_20,g_21,g_22,g_23], [g_30,g_31,g_32,g_33]]).subs(self.r, self.r_sub)
        # self.g_mu_nu_cart = self.g_mu_nu_cart.subs(self.alp, self.alp_sub).subs(self.r, self.r_sub)

        # self.g__mu__nu_cart_diff = [self.g__mu__nu_cart.diff(self.t), self.g__mu__nu_cart.diff(self.x), \
        #                              self.g__mu__nu_cart.diff(self.y), self.g__mu__nu_cart.diff(self.z)]

        # # We lambdify these to get numpy arrays
        # self.g__mu__nu_cart_lamb = sp.lambdify([self.t, self.x, self.y, self.z, self.r_s], self.g__mu__nu_cart, "numpy")
        # self.g_mu_nu_cart_lamb = sp.lambdify([self.t, self.x, self.y, self.z, self.r_s], self.g_mu_nu_cart, "numpy")
        # self.g__mu__nu_cart_diff_lamb = sp.lambdify([self.t, self.x, self.y, self.z, self.r_s], self.g__mu__nu_cart_diff, "numpy")

        # # Norm of k
        # # the norm of k determines if you have a massive particle (-1), a mass-less photon (0) 
        # # or a space-like curve (1)
        # self.k_t, self.k_x, self.k_y, self.k_z = sp.symbols('k_t k_x k_y k_z', real=True)
        # self.k_mu_cart = sp.Matrix([self.k_t, self.k_x, self.k_y, self.k_z])
        # self.norm_k = (self.k_mu_cart.T*self.g__mu__nu_cart*self.k_mu_cart)[0]

        # self.norm_k_lamb = sp.lambdify([self.k_t, self.k_x, self.k_y, self.k_z, self.x, self.y, self.z, \
        #                                        self.r_s], self.norm_k, "numpy")


    # ################################################################################################
    # #
    # ################################################################################################
    # def get_dk(self, kt_val, kx_val, ky_val, kz_val, t_val, x_val, y_val, z_val):
    #     # Calc g, g_inv and g_diff at given coords
    #     g = self.g__mu__nu_cart_lamb(t_val, x_val, y_val, z_val, self.r_s_value)
    #     g_inv = self.g_mu_nu_cart_lamb(t_val, x_val, y_val, z_val, self.r_s_value)
    #     g_diff = self.g__mu__nu_cart_diff_lamb(t_val, x_val, y_val, z_val, self.r_s_value)

    #     # Calc the connection Symbols at given coords
    #     gam_t = np.array([[self.gamma_func(g, g_inv, g_diff, 0, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
    #     gam_x = np.array([[self.gamma_func(g, g_inv, g_diff, 1, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
    #     gam_y = np.array([[self.gamma_func(g, g_inv, g_diff, 2, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
    #     gam_z = np.array([[self.gamma_func(g, g_inv, g_diff, 3, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])

    #     # Building up the geodesic equation: 
    #     # Derivatives: k_beta = d x^beta / d lambda
    #     #self.k_t, self.k_x, self.k_y, self.k_z = sp.symbols('k_t k_x k_y k_z', real=True)
    #     #self.k = [self.k_t, self.k_x, self.k_y, self.k_z]
    #     k = [kt_val, kx_val, ky_val, kz_val]
    
    #     # Second derivatives: d k_beta = d^2 x^beta / d lambda^2
    #     dk_t = np.sum(np.array([- gam_t[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]]))
    #     dk_x = np.sum(np.array([- gam_x[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]]))
    #     dk_y = np.sum(np.array([- gam_y[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]]))
    #     dk_z = np.sum(np.array([- gam_z[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]]))

    #     return dk_t, dk_x, dk_y, dk_z

    # # Connection Symbols
    # def gamma_func(self, g, g_inv, g_diff, sigma, mu, nu):
    #     # 29 Jan 2025
    #     # A list comprehension does NOT speed the following code up!
    #     # This something like this is NOT better:
    #     # g_sigma_mu_nu = np.sum(np.array( [ 1/2 * g_inv[sigma, rho] * (g_diff[mu][nu, rho] + g_diff[nu][rho, mu] - g_diff[rho][mu, nu] ) for rho in [0,1,2,3]] ) )
    #     # This is because you make a list and then sum, while you do not need to make the list. Normal for loop is fasters that I found so far.
    #     g_sigma_mu_nu = 0
    #     for rho in [0,1,2,3]:
    #         g_sigma_mu_nu += 1/2 * g_inv[sigma, rho] * (\
    #                         g_diff[mu][nu, rho] + \
    #                         g_diff[nu][rho, mu] - \
    #                         g_diff[rho][mu, nu] )
    #     return g_sigma_mu_nu

    # def k_t_from_norm(self, k0, x0, t=0):
    #     # Now we calculate k_t using the norm. This eliminates one of the differential equations.
    #     # time_like = True: calculates a geodesic for a massive particle
    #     # time_like = False: calculates a geodesic for a photon
    #     if (self.time_like):
    #         def wrap(k_t): return self.norm_k_lamb(k_t, k0[0], k0[1], k0[2], x0[0], x0[1], x0[2], self.r_s_value)+1
    #         k_t_from_norm = fsolve(wrap, 1.0)
    #     else:
    #         k_t_from_norm = root = fsolve(self.norm_k_lamb, 1.0, args = (k0[0], k0[1], k0[2], x0[0], x0[1], x0[2], self.r_s_value) )

    #     return k_t_from_norm[0]

    ################################################################################################
    #
    ################################################################################################
    def calc_trajectory(self, k0_xyz, x0_xyz, *args, **kargs):

        if not isinstance(k0_xyz, np.ndarray): k0_xyz = np.array(k0_xyz)
        if not isinstance(x0_xyz, np.ndarray): x0_xyz = np.array(x0_xyz)


        if k0_xyz.shape != x0_xyz.shape:
            print("k and x are not the same shape")
            return

        if k0_xyz.ndim == 1:
            if k0_xyz.shape[0] != 3 or x0_xyz.shape[0] != 3:
                print("k or x do not have 3 components")
                return
            k0_xyz = k0_xyz.reshape(1,3)
            x0_xyz = x0_xyz.reshape(1,3)

        else:
            if k0_xyz.shape[1] != 3 or x0_xyz.shape[1] != 3:
                print("k or x do not have 3 components")
                return

        if len(k0_xyz) == 1:
            return self.calc_trajectory_xyz(k0_xyz[0], x0_xyz[0], *args, **kargs)
        else:
            return [self.calc_trajectory_xyz(k0_xyz[i], x0_xyz[i], *args, **kargs) for i in range(len(x0_xyz))]

    ################################################################################################
    # calc_trajectory
    ################################################################################################
    # This function does the numerical integration of the geodesic equation using scipy's solve_ivp
    def calc_trajectory_xyz(self, \
                        k0_xyz = np.array([1, 0.0, 0.0]), x0_xyz = np.array([-10, 10, 0]), \
                        R_end = -1,\
                        curve_start = 0, \
                        curve_end = 50, \
                        nr_points_curve = 50, \
                        method = "RK45",\
                        max_step = np.inf,\
                        first_step = None,\
                        rtol = 1e-3,\
                        atol = 1e-6,\
                        verbose = False \
                       ):

        # Calculate from norm of starting condition
        k0_t = self.metric.k_t_from_norm(k0=k0_xyz, x0= x0_xyz, t = 0)
        values_0 = [k0_t, *k0_xyz, *x0_xyz]


        r0 = np.linalg.norm(x0_xyz)
        if R_end == -1:
            R_end = np.inf
        elif R_end < r0:
            R_end = np.sqrt(x0**2 + y0**2 + z0**2)*1.01

        if r0 > self.r_s_value:
            # Step function needed for solve_ivp
            def step(lamb, new):
                k_t_new, k_x_new, k_y_new, k_z_new, x_new, y_new, z_new = new

                dk_t, dk_x, dk_y, dk_z = self.metric.get_dk(k_t_new, k_x_new, k_y_new, k_z_new, \
                                                    t_val=lamb, x_val=x_new, y_val=y_new, z_val=z_new)

                dx, dy, dz = k_x_new, k_y_new, k_z_new
                return( dk_t, dk_x, dk_y, dk_z, dx, dy, dz)

            def hit_blackhole(t, y): 
                eps = 0.01
                k_t, k_x, k_y, k_z, x, y, z = y
                return np.sqrt(x**2+y**2+z**2) - (self.r_s_value+eps)
            hit_blackhole.terminal = True

            def reached_end(t, y): 
                k_t, k_x, k_y, k_z, x, y, z = y
                return np.sqrt(x**2+y**2+z**2) - R_end
            reached_end.terminal = True
            
            if nr_points_curve == 0:
                t_pts = None
            else:
                t_pts = np.linspace(curve_start, curve_end, nr_points_curve)

            start = time.time()
            events = [hit_blackhole]
            events.append(reached_end)

            result = solve_ivp(step, (curve_start, curve_end), values_0, t_eval=t_pts, \
                               events=events,\
                               method=method,\
                               max_step = max_step,\
                               first_step = first_step,\
                               atol=atol,\
                               rtol = rtol)
            end = time.time()
            if verbose: print("New: ", result.message, end-start, "sec")

            result.update({"R_end": R_end})
            result.update({"hit_blackhole": len(result.t_events[0])>0})
            result.update({"start_inside_hole": False})
            result.update({"end_check": len(result.t_events[1])>0})

            k_t, k_x, k_y, k_z, x, y, z = result.y
            t = result.t

            k3 = np.array([k_x, k_y, k_z])
            k4 = np.array([k_t, k_x, k_y, k_z])
            x3 = np.array([x, y, z])
            x4 = np.array([t, x, y, z])
            
            result.update({"k4_xyz": k4, "x4_xyz": x4})
            result.update({"k3_xyz": k3, "x3_xyz": x3})

            return k3, x3, result

        else:
            if verbose: print("Starting location inside the blackhole.")
            result = {"start_inside_hole": True}

            return result




import unittest
import curvedpy as cp
import sympy as sp
import numpy as np
import random
from curvedpy.utils.conversions import Conversions
from curvedpy.cameras.camera import RelativisticCamera


# python -m unittest discover -v test
# python test/test_geodesic_integration.py -v

################################################################################################
#
################################################################################################
class TestCameras(unittest.TestCase):
    def setUp(self):
        self.converter = Conversions()


    def test_CAM_make_camera(self):
        self.camera = RelativisticCamera(resolution = [2, 2], verbose_init = False, verbose=False)
        self.camera.run_mp(cores=1)

        self.assertTrue( len(self.camera.results) == 4 )
        self.assertTrue( len(self.camera.geodesics) == 4 )
        self.assertTrue( len(self.camera.pixel_coordinates) == 4 )


        self.assertTrue( self.camera.ray_blackhole_hit.shape == (2,2,1) )
        #self.assertTrue( self.camera.ray_end.shape == (2,2,6) ) 

if __name__ == '__main__':
    unittest.main()




#import sympy as sp

import numpy as np  # Thinly-wrapped numpy
import sympy as sp

#from autograd import grad #
#from autograd import elementwise_grad as egrad  # for functions that vectorize over inputs
#from autograd import jacobian  # for functions that vectorize over inputs

from curvedpy.utils.stringified_equations import KerrSchildStrings
from sympy.parsing.sympy_parser import parse_expr

from curvedpy.utils.conversions_4D import Conversions4D
from scipy.optimize import fsolve


# !!
# Kerr Schild Metrics implemented using Numpy and Autograd
# !!
# Author: B.L. de Vries

################################################################################################
################################################################################################
class KerrSchildMetricXYZ:
################################################################################################
################################################################################################

    conversions4D = Conversions4D()

    ################################################################################################
    # INIT
    ################################################################################################
    def __init__(self, mass=1.0, a = 0.1, time_like = False, verbose=False):
        self.mass = mass
        self.a = a

        # Type of geodesic
        self.time_like = time_like # No Massive particle geodesics yet
        self.verbose = verbose

        t, x, y, z, r, a, m = sp.symbols('t x y z r a m')
        R2 = x**2 + y**2 + z**2
        r_sub = sp.sqrt( (R2 - a**2 + sp.sqrt((R2 - a**2)**2 +4*a**2*z**2 ) )/2 )

        self.r_lamb = sp.lambdify([x, y, z, a], r_sub, 'numpy')
        self.g__mu__nu = sp.lambdify([t, x, y, z, r, a, m], parse_expr(KerrSchildStrings.str_g__mu__nu), 'numpy')
        self.g_mu_nu = sp.lambdify([t, x, y, z, r, a, m], parse_expr(KerrSchildStrings.str_g_mu_nu), 'numpy')
        self.g__mu__nu_diff = [sp.lambdify([t, x, y, z, r, a, m], parse_expr(KerrSchildStrings.str_g__mu__nu_diff[i]), 'numpy') \
                                for i in [0,1,2,3]]


        # Outer horizon in BL coordinate r
        self.r_plus = self.mass + np.sqrt(self.mass**2 - self.a**2) # See Visser 2008, page 24, eq 104

    ################################################################################################
    #
    ################################################################################################
    def check_horizin(self, x, y, z):
        # See Visser 2008 page 28 (eq 126)
        return x**2 + y**2 + (2*self.mass/self.r_plus) * z**2 - 2*self.mass*self.r_plus

    ################################################################################################
    #
    ################################################################################################
    def get_dk(self, kt_val, kx_val, ky_val, kz_val, t_val, x_val, y_val, z_val):
        # Calc g_inv and g_diff at given coords
        r_val = self.r_lamb(x_val, y_val, z_val, self.a)
        g_inv = self.g_mu_nu(t_val, x_val, y_val, z_val, r_val, self.a, self.mass)
        g_diff = [self.g__mu__nu_diff[i](t_val, x_val, y_val, z_val, r_val, self.a, self.mass) for i in [0,1,2,3]]

        # Calc the connection Symbols at given coords
        gam_t = np.array([[self.gamma_func(g_inv, g_diff, 0, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        gam_x = np.array([[self.gamma_func(g_inv, g_diff, 1, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        gam_y = np.array([[self.gamma_func(g_inv, g_diff, 2, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        gam_z = np.array([[self.gamma_func(g_inv, g_diff, 3, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])

        # Building up the geodesic equation: 
        # Derivatives: k_beta = d x^beta / d lambda
        #self.k_t, self.k_x, self.k_y, self.k_z = sp.symbols('k_t k_x k_y k_z', real=True)
        #self.k = [self.k_t, self.k_x, self.k_y, self.k_z]
        k = [kt_val, kx_val, ky_val, kz_val]
    
        # Second derivatives: d k_beta = d^2 x^beta / d lambda^2
        dk_t = np.sum(np.array([- gam_t[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]]))
        dk_x = np.sum(np.array([- gam_x[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]]))
        dk_y = np.sum(np.array([- gam_y[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]]))
        dk_z = np.sum(np.array([- gam_z[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]]))

        return dk_t, dk_x, dk_y, dk_z

    ################################################################################################
    # Connection Symbols
    ################################################################################################
    def gamma_func(self, g_inv, g_diff, sigma, mu, nu):
        # 29 Jan 2025
        # A list comprehension does NOT speed the following code up!
        # Something like this is NOT better:
        # g_sigma_mu_nu = np.sum(np.array( [ 1/2 * g_inv[sigma, rho] * (g_diff[mu][nu, rho] \
        #               + g_diff[nu][rho, mu] - g_diff[rho][mu, nu] ) for rho in [0,1,2,3]] ) )
        # This is because you make a list and then sum, while you do not need to make the list.
        # I found that it is faster to use a normal for loop like this:
        g_sigma_mu_nu = 0
        for rho in [0,1,2,3]:
            g_sigma_mu_nu += 1/2 * g_inv[sigma, rho] * (\
                            g_diff[mu][nu, rho] + \
                            g_diff[nu][rho, mu] - \
                            g_diff[rho][mu, nu] )
        return g_sigma_mu_nu

    ################################################################################################
    # Norm of the four vector k
    ################################################################################################
    def norm_k(self, k4, x4):#k_t, k_x, k_y, k_z, t, x, y, z):
        # Norm of k
        # the norm of k determines if you have a massive particle (-1), a mass-less photon (0) 
        # or a space-like curve (1) 
        r_val = self.r_lamb(x4[1], x4[2], x4[3], self.a)

        return (k4.T @ self.g__mu__nu(*x4, r_val, self.a, self.mass) @ k4) #t, x, y, z, r, a, m

    ################################################################################################
    # Solve k_t for the starting 4-momentum 
    ################################################################################################
    def k_t_from_norm(self, k0, x0, t=0):
        # Now we calculate k_t using the norm. This eliminates one of the differential equations.
        # time_like = True: calculates a geodesic for a massive particle
        # time_like = False: calculates a geodesic for a photon

        # Starting coordinates in 4D
        x4_0 = np.array([t, *x0])

        if (self.time_like):
            def wrap(k_t):
                # Starting 4-vector
                k4_0 = np.concatenate((k_t, k0))
                return self.norm_k(k4_0, x4_0)+1

            k_t_from_norm = fsolve(wrap, 1.0)
        else:
            def wrap(k_t):
                # Starting 4-vector
                k4_0 = np.concatenate((k_t, k0))
                return self.norm_k(k4_0, x4_0)
                
            k_t_from_norm = fsolve(wrap, 1.0)

        return k_t_from_norm[0]







import unittest
import curvedpy as cp
import sympy as sp
import numpy as np
import random
from curvedpy.utils.conversions import Conversions
from curvedpy.geodesics.blackhole import BlackholeGeodesicIntegrator
from curvedpy.metrics.schwarzschild_metric import SchwarzschildMetricSpherical
from curvedpy.utils.conversions_SPH2PATCH_3D import cart_to_sph1

# python -m unittest discover -v test
# python test/test_geodesic_integration.py -v
################################################################################################
# Test to see if the spherical to cartesian conversion functions are consistent
################################################################################################
class TestConversions(unittest.TestCase):
    def setUp(self):
        self.converter = Conversions()

    def test_SCHW_sph_to_xyz_and_back(self):
        k0_xyz = np.array([11.322145, 15.136237, 65.246265])
        x0_xyz = np.array([13.461341, 13.461346, 72.300564])
        round_lvl = 6

        x0_sph, k0_sph = self.converter.convert_xyz_to_sph(x0_xyz, k0_xyz)
        x0_xyz_new, k0_xyz_new = self.converter.convert_sph_to_xyz(x0_sph, k0_sph)

        self.assertTrue( bool((k0_xyz == [round(v, round_lvl) for v in k0_xyz_new]).all()) )
        self.assertTrue( bool((x0_xyz == [round(v, round_lvl) for v in x0_xyz_new]).all()) )







################################################################################################
# Test to check if the geodesic is the same in the other direction. This means the 
# metric is not time dependent.
################################################################################################

class TestCurvedpySchwarzschild(unittest.TestCase):

    def setUp(self):
        self.gi = BlackholeGeodesicIntegrator()
        #self.gi = cp.GeodesicIntegratorSchwarzschild()#metric='schwarzschild', mass=1.0)
        self.start_t, self.end_t, self.steps = 0, 60, 60
        self.max_step = 1
        self.round_level = 4

    def test_SCHW_check_direction_symmetry(self):  
        # Check if you get the same line forward or backward

        # Line in forward direction
        k0_xyz = np.array([1, 0.0, 0.0])
        x0_xyz = np.array([-10, 10, 0])

        k_xyz, x_xyz, line = self.gi.geodesic(k0_xyz, x0_xyz, \
                                  curve_start = self.start_t, \
                                  curve_end = self.end_t, \
                                  nr_points_curve = self.steps,\
                                 max_step=self.max_step)

        k_end = sp.Matrix([k_xyz[0][-1], k_xyz[1][-1], k_xyz[2][-1]])

        # Rotation matrix for 180 degrees rotation around the z axis
        th = sp.Symbol("\\theta")
        R90 = sp.rot_axis3(th).subs(th, sp.pi)
        k_0_back = R90*k_end

        k0_xyz2 = np.array(k_0_back).astype(np.float64).flatten() #
        x0_xyz2 = np.array([x_xyz[0][-1], x_xyz[1][-1], x_xyz[2][-1]])


        # Trajectory in backward direction
        k_xyz2, x_xyz2, line_reverse = self.gi.geodesic(k0_xyz2, x0_xyz2, \
                                  curve_start = self.start_t, \
                                  curve_end = self.end_t, \
                                  nr_points_curve = self.steps,\
                                 max_step=self.max_step)

        self.assertTrue( bool((np.round(x_xyz[0],self.round_level) == np.round(np.flip(x_xyz2[0]),self.round_level)).all()) )
        self.assertTrue( bool((np.round(x_xyz[1],self.round_level) == np.round(np.flip(x_xyz2[1]),self.round_level)).all()) )
        self.assertTrue( bool((np.round(x_xyz[2],self.round_level) == np.round(np.flip(x_xyz2[2]),self.round_level)).all()) )




    # def test_check_constant_kt(self):
    #     k0_xyz = np.array([1, 0.0, 0.0])
    #     x0_xyz = np.array([-10, 10, 0])

    #     k_xyz, x_xyz, line = self.gi.calc_trajectory(k0_xyz, x0_xyz, \
    #                               curve_start = self.start_t, \
    #                               curve_end = self.end_t, \
    #                               nr_points_curve = self.steps,\
    #                              max_step=self.max_step)
    
    #     k_r, r, k_th, th, k_ph, ph, k_t = line.y
    #     k_t = self.gi.k_t_from_norm_lamb(k_r, r, k_th, th, k_ph, ph, self.gi.r_s_value)

    #     self.assertTrue( bool(np.std(k_t) < 0.077) ) # Could be something to improve?

    #     # Also check if the norm of a null ray in nicely zero
    #     norm_k = self.gi.norm_k_lamb(k_t, k_r, r, k_th, th, k_ph, ph, self.gi.r_s_value)
    #     self.assertTrue( round(np.std(norm_k),8) == 0.0 )


################################################################################################
# Testing if the energy and angular momentum is conserved for orbits of photons or massive particles.
# This test is similar to what Bronzwaer et al. 18 and Bronzwaer et al. 21
################################################################################################
class TestCurvedpySchwarzschild_conservation(unittest.TestCase):
    def setUp(self):
        self.converter = Conversions()

        self.mass = 1.0
        self.start_t, self.end_t, self.steps = 0, 60, 60
        self.max_step = 0.1
        self.round_level = 6

        self.metric_sph = SchwarzschildMetricSpherical(mass = self.mass)


    def test_SCHW_SPH_check_conserved_quantities_photons(self):
        self.gi = BlackholeGeodesicIntegrator(mass = self.mass, coordinates="SPH2PATCH", time_like = False)

        k0_sph = np.array([0.0, 0., -0.1]) 
        x0_sph = np.array([3, 1/2*np.pi, 1/4*np.pi])

        x0_xyz, k0_xyz = self.converter.convert_sph_to_xyz(x0_sph, k0_sph)
        k, x, res =  self.gi.geodesic(k0_xyz, x0_xyz, verbose=False, max_step=self.max_step)#curve_end = 100, nr_points_curve = 1000, 

        k_sph, x_sph = cart_to_sph1(k, x)
        k4 = np.array([res['k_t'], *k_sph])
        x4 = np.array([res['t'], *x_sph])
        k4__mu = self.gi.get_metric().oneform(k4, x4)
        L = k4__mu[3]
        E = k4__mu[0]

        self.assertTrue( round(np.std(L),self.round_level) == 0.0 )
        self.assertTrue( round(np.std(E),self.round_level) == 0.0 )

    def test_SCHW_SPH_check_conserved_quantities_massive_particles(self):
        self.gi = BlackholeGeodesicIntegrator(mass = self.mass, coordinates="SPH2PATCH", time_like = True)

        #self.gi = cp.GeodesicIntegratorSchwarzschild(mass = self.mass, time_like = True)#metric='schwarzschild', mass=1.0)

        k0_sph = np.array([0., 0., -0.1])
        x0_sph = np.array([20, 1/2*np.pi,0])
        x0_xyz, k0_xyz = self.converter.convert_sph_to_xyz(x0_sph, k0_sph)

        k, x, res =  self.gi.geodesic(k0_xyz, x0_xyz, verbose=False, max_step=self.max_step)#curve_end = 100, nr_points_curve = 1000, 
        k_sph, x_sph = cart_to_sph1(k, x)
        k4 = np.array([res['k_t'], *k_sph])
        x4 = np.array([res['t'], *x_sph])
        k4__mu = self.gi.get_metric().oneform(k4, x4)

        #k_r, r, k_th, th, k_ph, ph, k_t = res.y

        # Get the four vectors
        # k4_mu_xyz, x4_mu_xyz = res.k4_xyz, res.x4_xyz
        # x3_mu_sph, k3_mu_sph = self.converter.convert_xyz_to_sph(x4_mu_xyz[1:], k4_mu_xyz[1:])
        # x4_mu_sph = np.array([x4_mu_xyz[0], *x3_mu_sph])
        # k4_mu_sph = np.array([k4_mu_xyz[0], *k3_mu_sph])
        # # Convert the 4vectors to oneforms
        # k__mu = self.ssm.oneform(k4_mu_sph, x4_mu_sph)

        L = k4__mu[3]
        self.assertTrue( round(np.std(L),self.round_level) == 0.0 )
        E = k4__mu[0]
        self.assertTrue( round(np.std(E),self.round_level) == 0.0 )

    def test_SCHW_XYZ_check_conserved_quantities_photons(self):
        self.gi = BlackholeGeodesicIntegrator(mass = self.mass, coordinates="xyz", time_like = False)

        k0_sph = np.array([0.0, 0., -0.1]) 
        x0_sph = np.array([3, 1/2*np.pi, 1/4*np.pi])
        x0_xyz, k0_xyz = self.converter.convert_sph_to_xyz(x0_sph, k0_sph)
        k, x, res =  self.gi.geodesic(k0_xyz, x0_xyz, verbose=False, max_step=self.max_step)
        k_t, k_x, k_y, k_z, _,_,_ = res.y
        t = res.t

        k_sph, x_sph = cart_to_sph1(k, x)
        k4 = np.array([k_t, *k_sph])
        x4 = np.array([t, *x_sph])
        k4__mu = self.metric_sph.oneform(k4, x4)

        L = k4__mu[3]
        self.assertTrue( round(np.std(L),self.round_level) == 0.0 )
        E = k4__mu[0]
        self.assertTrue( round(np.std(E),self.round_level) == 0.0 )


if __name__ == '__main__':
    unittest.main()



import importlib.metadata
__version__ = importlib.metadata.version('curvedpy')

from curvedpy.geodesics.blackhole import BlackholeGeodesicIntegrator
from curvedpy.cameras.camera import RelativisticCamera
from curvedpy.utils.projections import projection


import sympy as sp
import numpy as np
from scipy.integrate import solve_ivp
import time


class Integrator4D:

    def __init__(self):
        pass

    ################################################################################################
    # 
    ################################################################################################
    # This function does the numerical integration of the geodesic equation using scipy's solve_ivp
    def integrate(self, \
                        k4_start, x4_start, \
                        get_dk,\
                        hit_blackhole,\
                        stop_integration = None,\
                        stop_integration_coord_check = None,\
                        curve_start = 0, \
                        curve_end = 50, \
                        nr_points_curve = 50, \
                        method = "RK45",\
                        max_step = np.inf,\
                        first_step = None,\
                        rtol = 1e-3,\
                        atol = 1e-6,\
                        verbose = False \
                       ):
        
        # Step function needed for solve_ivp
        def step(lamb, new):
            k_0_new, k_1_new, k_2_new, k_3_new, x_0_new, x_1_new, x_2_new, x_3_new = new

            dk_0, dk_1, dk_2, dk_3 = get_dk(k_0_new, k_1_new, k_2_new, k_3_new, x_0_new, x_1_new, x_2_new, x_3_new)
            dx_0, dx_1, dx_2, dx_3 = k_0_new, k_1_new, k_2_new, k_3_new

            return( dk_0, dk_1, dk_2, dk_3, dx_0, dx_1, dx_2, dx_3)

        # EVENTS
        # This is not perfectly general yet!!
        events = []

        hit_blackhole.terminal = True
        events.append(hit_blackhole)

        if stop_integration:
            stop_integration.terminal = True
            stop_integration.direction = +1
            events.append(stop_integration)

        if stop_integration_coord_check:
            stop_integration_coord_check.terminal = True
            events.append(stop_integration_coord_check)

        values_0 = [ *k4_start, *x4_start ]

        if nr_points_curve == 0:
            t_pts = None
        else:
            t_pts = np.linspace(curve_start, curve_end, nr_points_curve)

        start = time.time()
        result = solve_ivp(step, (curve_start, curve_end), values_0, t_eval=t_pts, \
                           events=events,\
                           method=method,\
                           max_step = max_step,\
                           first_step = first_step,\
                           atol=atol,\
                           rtol = rtol)
        end = time.time()
        if verbose: print("New: ", result.message, end-start, "sec")

        result.update({"hit_blackhole": len(result.t_events[0])>0})
        if stop_integration:
            result.update({"end_check": len(result.t_events[1])>0})
        if stop_integration_coord_check:
            result.update({"stop_integration_coord_check": len(result.t_events[2])>0})

        return result


import sympy as sp
import numpy as np
from scipy.integrate import solve_ivp
import time
from curvedpy.metrics.kerrschild_metric import KerrSchildMetricXYZ
#from curvedpy.metrics.schwarzschild_metric_AUTOGRAD import SchwarzschildMetricXYZ_AUTOGRAD

from curvedpy.utils.conversions import Conversions

# -------------------------------
# ----NAMING CONVENTIONS USED----
# -------------------------------
#
# UPPER & LOWER INDICES
# _ indicates an upper/contravariant index
# __ indicates a lower/covariant index
# For example g__mu__nu is the metric g with two lower indices while g_mu_nu is the metric 
# with two upper indices, which is the inverse of g__mu__nu
# Another example is 4-momentum k_mu, which has a 0 (time) component of k_0 or k_t
# The 4-momentum as a oneform is then k__mu and its zeroth component is k__0
# An example tensor could be T_mu__nu__alp. This tensor has one upper/contravariant (mu) and 
# two lower/covariant/oneform (nu, alp) indices.
#
# COORDINATES
# the used coordinate system for a tensor is indicated with by appending for example 
# _xyz, _sph, _bl after the indices.
# Example: x_mu_bl: this 4vector with one upper index is given in Boyer-Lindquist coordinates
# Example: g__mu__nu_sph: this metric tensor with two lower indices is given in spherical coordinates
#
# MISC
# If a vector x_mu has only 3 components, they are the three spatial components


# https://f.yukterez.net/einstein.equations/files/8.html#transformation
# https://physics.stackexchange.com/questions/672252/how-to-compute-and-interpret-schwartzchild-black-hole-metric-line-element-ds-i
class GeodesicIntegratorKerrSchildXYZ:

    conversions = Conversions()

    ################################################################################################
    #
    ################################################################################################
    def __init__(self, mass=1.0, a = 0.1, time_like = False, verbose=False):

        self.M = mass
        self.a = a

        # Type of geodesic
        self.time_like = time_like # No Massive particle geodesics yet
        self.verbose = verbose
        if verbose:
            print("Geodesic SS Integrator Settings: ")
            print(f"  - {self.M=}")
            print(f"  - {self.a=}")
            print(f"  - {self.time_like=}")
            print(f"  - {self.verbose=}")
            print("--")

        self.metric = KerrSchildMetricXYZ(mass, a, time_like, verbose)

    ################################################################################################
    #
    ################################################################################################
    def calc_trajectory(self, k0_xyz, x0_xyz, *args, **kargs):

        if not isinstance(k0_xyz, np.ndarray): k0_xyz = np.array(k0_xyz)
        if not isinstance(x0_xyz, np.ndarray): x0_xyz = np.array(x0_xyz)


        if k0_xyz.shape != x0_xyz.shape:
            print("k and x are not the same shape")
            return

        if k0_xyz.ndim == 1:
            if k0_xyz.shape[0] != 3 or x0_xyz.shape[0] != 3:
                print("k or x do not have 3 components")
                return
            k0_xyz = k0_xyz.reshape(1,3)
            x0_xyz = x0_xyz.reshape(1,3)

        else:
            if k0_xyz.shape[1] != 3 or x0_xyz.shape[1] != 3:
                print("k or x do not have 3 components")
                return

        if len(k0_xyz) == 1:
            return self.calc_trajectory_xyz(k0_xyz[0], x0_xyz[0], *args, **kargs)
        else:
            return [self.calc_trajectory_xyz(k0_xyz[i], x0_xyz[i], *args, **kargs) for i in range(len(x0_xyz))]

    ################################################################################################
    # calc_trajectory
    ################################################################################################
    # This function does the numerical integration of the geodesic equation using scipy's solve_ivp
    def calc_trajectory_xyz(self, \
                        k0_xyz = np.array([1, 0.0, 0.0]), x0_xyz = np.array([-10, 10, 0]), \
                        R_end = -1,\
                        curve_start = 0, \
                        curve_end = 50, \
                        nr_points_curve = 50, \
                        method = "RK45",\
                        max_step = np.inf,\
                        first_step = None,\
                        rtol = 1e-3,\
                        atol = 1e-6,\
                        verbose = False, *args, **kargs \
                       ):

        # Calculate from norm of starting condition
        k0_t = self.metric.k_t_from_norm(k0=k0_xyz, x0= x0_xyz, t = 0)
        values_0 = [k0_t, *k0_xyz, *x0_xyz]


        r0 = np.linalg.norm(x0_xyz)
        if R_end == -1:
            R_end = np.inf
        elif R_end < r0:
            R_end = np.sqrt(x0**2 + y0**2 + z0**2)*1.01

        if self.metric.check_horizin(*x0_xyz) > 0.0:
            # Step function needed for solve_ivp
            def step(lamb, new):
                k_t_new, k_x_new, k_y_new, k_z_new, x_new, y_new, z_new = new

                dk_t, dk_x, dk_y, dk_z = self.metric.get_dk(k_t_new, k_x_new, k_y_new, k_z_new, \
                                                    t_val=lamb, x_val=x_new, y_val=y_new, z_val=z_new)

                dx, dy, dz = k_x_new, k_y_new, k_z_new
                return( dk_t, dk_x, dk_y, dk_z, dx, dy, dz)

            def hit_blackhole(t, y): 
                eps = 0.01
                k_t, k_x, k_y, k_z, x, y, z = y
                return self.metric.check_horizin(x, y, z)
                #return np.sqrt(x**2+y**2+z**2) - (self.r_s_value+eps)
            hit_blackhole.terminal = True

            def reached_end(t, y): 
                k_t, k_x, k_y, k_z, x, y, z = y
                return np.sqrt(x**2+y**2+z**2) - R_end
            reached_end.terminal = True
            
            if nr_points_curve == 0:
                t_pts = None
            else:
                t_pts = np.linspace(curve_start, curve_end, nr_points_curve)

            start = time.time()
            events = [hit_blackhole]
            events.append(reached_end)

            result = solve_ivp(step, (curve_start, curve_end), values_0, t_eval=t_pts, \
                               events=events,\
                               method=method,\
                               max_step = max_step,\
                               first_step = first_step,\
                               atol=atol,\
                               rtol = rtol)
            end = time.time()
            if verbose: print("New: ", result.message, end-start, "sec")

            result.update({"R_end": R_end})
            result.update({"hit_blackhole": len(result.t_events[0])>0})
            result.update({"start_inside_hole": False})
            result.update({"end_check": len(result.t_events[1])>0})

            k_t, k_x, k_y, k_z, x, y, z = result.y
            t = result.t

            k3 = np.array([k_x, k_y, k_z])
            k4 = np.array([k_t, k_x, k_y, k_z])
            x3 = np.array([x, y, z])
            x4 = np.array([t, x, y, z])
            
            result.update({"k4_xyz": k4, "x4_xyz": x4})
            result.update({"k3_xyz": k3, "x3_xyz": x3})

            return k3, x3, result

        else:
            if verbose: print("Starting location inside the blackhole.")
            result = {"start_inside_hole": True}

            return result




import sympy as sp
import numpy as np
#from scipy.integrate import solve_ivp
from scipy.optimize import fsolve
import time
from curvedpy.utils.conversions import Conversions
from curvedpy.geodesics.blackhole_integrators.integrator_4D import Integrator4D
from curvedpy.metrics.schwarzschild_metric import SchwarzschildMetricSpherical

from curvedpy.utils.conversions_SPH2PATCH_3D import cart_to_sph2, cart_to_sph1, sph1_to_cart, sph2_to_cart, sph1_to_sph2, sph2_to_sph1, coords_sph1_to_cart, coords_sph2_to_cart

# -------------------------------
# ----NAMING CONVENTIONS USED----
# -------------------------------
#
# UPPER & LOWER INDICES
# _ indicates an upper/contravariant index
# __ indicates a lower/covariant index
# For example g__mu__nu is the metric g with two lower indices while g_mu_nu is the metric 
# with two upper indices, which is the inverse of g__mu__nu
# Another example is 4-momentum k_mu, which has a 0 (time) component of k_0 or k_t
# The 4-momentum as a oneform is then k__mu and its zeroth component is k__0
# An example tensor could be T_mu__nu__alp. This tensor has one upper/contravariant (mu) and 
# two lower/covariant/oneform (nu, alp) indices.
#
# COORDINATES
# the used coordinate system for a tensor is indicated with by appending for example 
# _xyz, _sph, _bl after the indices.
# Example: x_mu_bl: this 4vector with one upper index is given in Boyer-Lindquist coordinates
# Example: g__mu__nu_sph: this metric tensor with two lower indices is given in spherical coordinates
#
# MISC
# If a vector x_mu has only 3 components, they are the three spatial components


# https://f.yukterez.net/einstein.equations/files/8.html#transformation
# https://physics.stackexchange.com/questions/672252/how-to-compute-and-interpret-schwartzchild-black-hole-metric-line-element-ds-i
class GeodesicIntegratorSchwarzschildSPH1SPH2:

    conversions = Conversions()
    integrator = Integrator4D()

    ################################################################################################
    #
    ################################################################################################
    def __init__(self, mass=1.0, time_like = False, theta_switch = 0.1*np.pi, eps_theta=0.0000001, verbose=False):
        self.M = mass
        self.theta_switch = theta_switch
        self.metric = SchwarzschildMetricSpherical(mass=mass, eps_theta=eps_theta)
        self.verbose = verbose

    ################################################################################################
    #
    ################################################################################################
    # This function checks the shape of the parameters and if multiple photons need to be run
    def calc_trajectory(self, k0_xyz, x0_xyz, full_save = False, *args, **kargs):

        if not isinstance(k0_xyz, np.ndarray): k0_xyz = np.array(k0_xyz)
        if not isinstance(x0_xyz, np.ndarray): x0_xyz = np.array(x0_xyz)


        if k0_xyz.shape != x0_xyz.shape:
            print("k and x are not the same shape")
            return

        if k0_xyz.ndim == 1:
            if k0_xyz.shape[0] != 3 or x0_xyz.shape[0] != 3:
                print("k or x do not have 3 components")
                return
            k0_xyz = k0_xyz.reshape(1,3)
            x0_xyz = x0_xyz.reshape(1,3)

        else:
            if k0_xyz.shape[1] != 3 or x0_xyz.shape[1] != 3:
                print("k or x do not have 3 components")
                return

        if len(k0_xyz) == 1:
            return self.calc_trajectory_2patch(k0_xyz[0], x0_xyz[0], full_save, *args, **kargs)
        else:
            return [self.calc_trajectory_2patch(k0_xyz[i], x0_xyz[i], full_save, *args, **kargs) for i in range(len(x0_xyz))]

    
    ################################################################################################
    # 
    ################################################################################################
    # This function handles the switching between the different coordinate patches
    def calc_trajectory_2patch(self, k0_xyz, x0_xyz, full_save = False, *args, **kargs):
        if not isinstance(x0_xyz,np.ndarray):
            x0_xyz = np.array(x0_xyz)

        if not isinstance(k0_xyz,np.ndarray):
            k0_xyz = np.array(k0_xyz)

        summary = {"coords":[]}

        # We start by assuming spherical coordinate patch 1
        # when theta is measured from the z-axis
        k0_sph, x0_sph = cart_to_sph1(k0_xyz, x0_xyz)
        coord = "sph1"
        if x0_sph[1] <= self.theta_switch:
            # Switch to second patch 2 if close to the theta-axis
            k0_sph, x0_sph = cart_to_sph2(k0_xyz, x0_xyz)
            coord = "sph2"
        summary["coords"].append(coord)

        # The geodesic pieces calculated in SPH1 or SPH2 will be stored in this list and
        # in cartesian coordinates.
        final_k_xyz, final_x_xyz = [], []
        # Here the propertime/parameter of the geodesic is tracked
        final_t = []
        t_end_prev_piece = 0.0
        final_k_t = []
        # The integrator information for every piece is tracked in this list
        results_list = []

        # Run the first integration for the first geodesic piece 
        if self.verbose: print(f"=> First run in {coord=}")
        k_sph_step, x_sph_step, res = self.calc_trajectory_sph(k0_sph, x0_sph, coord, *args, **kargs)
        k_t, k_r, k_th, k_ph, t, r, th, ph = res.y

        # Convert the first piece to cartesian and store it
        if coord == "sph1":
            k_xyz_step, x_xyz_step = sph1_to_cart(k_sph_step, x_sph_step)
        else: # coords == "sph2":
            k_xyz_step, x_xyz_step = sph2_to_cart(k_sph_step, x_sph_step)

        # NEED TO TRACK res.t! 
        final_t.append(np.array(t))#[1:]))#+t_end_prev_piece) # We do not need the first entry
        final_k_t.append(k_t)#[1:])

        # Save the used coordinate
        res.coord = coord
        # Save the geodesic piece
        res.x_xyz = x_xyz_step
        res.k_xyz = k_xyz_step
        # Add geodesic and result dict to final results
        results_list.append(res)
        final_k_xyz.append(np.column_stack(k_xyz_step))#[1:])
        final_x_xyz.append(np.column_stack(x_xyz_step))#[1:])

        # Now we will check if we need to integrate another piece or 
        # if we are done.
        nr_switches = 0
        while res.stop_integration_coord_check and (not res.hit_blackhole) and nr_switches <=4:
            # We start the next piece where the previous ended
            k_sph, x_sph = np.column_stack(k_sph_step), np.column_stack(x_sph_step)
            k0_sph, x0_sph = k_sph[-1], x_sph[-1]

            # We convert the end point of the previous piece to the 
            # other coordinate patch
            if coord == "sph1":
                k0_sph, x0_sph = sph1_to_sph2(k0_sph, x0_sph)
                coord = "sph2"
            else: # coord == "sph2"
                k0_sph, x0_sph = sph2_to_sph1(k0_sph, x0_sph)
                coord = "sph1"
            summary["coords"].append(coord)
            if self.verbose: print(f"=> Starting {nr_switches} ({res.stop_integration_coord_check}, {res.hit_blackhole})")
            if self.verbose: print(f" th0 = {round(x0_sph[1],2)} in {coord}")

            # Then we run the next piece
            k_sph_step, x_sph_step, res = self.calc_trajectory_sph(k0_sph, x0_sph, coord, use_t0_k_t_0 = True, t0 = t[-1], k_t_0 = k_t[-1], *args, **kargs)
            k_t, k_r, k_th, k_ph, t, r, th, ph = res.y

            # We convert the new piece to cartesian
            if coord == "sph1":
                k_xyz_step, x_xyz_step = sph1_to_cart(k_sph_step, x_sph_step)
            else: # coords == "sph2":
                k_xyz_step, x_xyz_step = sph2_to_cart(k_sph_step, x_sph_step)
            
            # Track the propertime/parameter and k_t
            final_t.append(t[1:])#+t_end_prev_piece) # We do not need the first entry
            final_k_t.append(k_t[1:])

            # We save all the results
            res.coord = coord
            res.x_xyz = x_xyz_step
            res.k_xyz = k_xyz_step
            results_list.append(res)
            final_k_xyz.append(np.column_stack(k_xyz_step)[1:])
            final_x_xyz.append(np.column_stack(x_xyz_step)[1:])

            # And increase the counter
            nr_switches += 1


        # Merge all pieces
        merged_k_xyz = np.column_stack(np.concatenate(final_k_xyz))
        merged_x_xyz = np.column_stack(np.concatenate(final_x_xyz))
        # Merge the propertime/parameter of the different pieces
        merged_t = np.concatenate(final_t)
        merged_k_t = np.concatenate(final_k_t)

        summary["t"] = merged_t
        summary["k_t"] = merged_k_t
        # If the last piece hits the blackhole the geodesic ends there
        summary["hit_blackhole"] = res["hit_blackhole"] 

        if full_save: 
            print(" == Carefull, doing full saving in schwarzschild_SPH1SPH2.py. This results in huge files for debugging == ")
            summary["result_list"] = results_list

        return merged_k_xyz, merged_x_xyz, summary


    ################################################################################################
    #
    ################################################################################################
    # This function calculates one piece of the geodesic in one coordinate patch
    def calc_trajectory_sph(self, \
                        k0_sph, x0_sph, \
                        coord,\
                        use_t0_k_t_0 = False, t0=0.0, k_t_0=0.0,\
                        R_end = 100,\
                        curve_start = 0, \
                        curve_end = 10000, \
                        nr_points_curve = 5000, \
                        method = "RK45",\
                        max_step = np.inf,\
                        first_step = None,\
                        rtol = 1e-3,\
                        atol = 1e-6,\
                        verbose = False \
                       ):

        k_r_0, k_th_0, k_ph_0 = k0_sph
        r0, th0, ph0 = x0_sph

        # Calculate from norm of starting condition
        if not use_t0_k_t_0:
            t0 = 0
            k_t_0 = self.metric.k_t_from_norm_lamb(k_r_0, k_th_0, k_ph_0, t0, r0, th0, ph0)

        k_0 = np.array([k_t_0, *k0_sph])
        x_0 = np.array([t0, *x0_sph])

        #Check if starting values are outside the blackhole
        if r0 < self.metric.get_r_s():
            print("Starting value inside blackhole")
            return

        def hit_blackhole(t, y): 
            eps = 0.01
            k_0, k_1, k_2, k_3, x_0, x_1, x_2, x_3 = y
            r = x_1#calc_radius_from_x_mu(x_0, x_1, x_2, x_3)
            return r - (self.metric.get_r_s()+eps)

        if R_end == -1: R_end = np.inf
        #elif R_end < r0: R_end = r0*1.01
        def stop_integration(t, y): 
            k_0, k_1, k_2, k_3, x_0, x_1, x_2, x_3 = y
            r = x_1 #calc_radius_from_x_mu(x_0, x_1, x_2, x_3)
            return r - R_end

        def switch_coordinates(t, y):
            if self.theta_switch < 0.0:
                return 1
            else:
                k_0, k_1, k_2, k_3, x_0, x_1, x_2, x_3 = y
                if coord == "sph1":
                    x, y, z = coords_sph1_to_cart(x_1, x_2, x_3)
                    th = np.acos(z/np.sqrt(x**2+y**2+z**2))
                    #print("sph1", th/np.pi)
                    # if abs(th) <= self.theta_switch:# and abs(th) >= 2*np.pi-self.theta_switch:
                    #     return 0
                    # else:
                    #     return 1
                else:
                    x, y, z = coords_sph2_to_cart(x_1, x_2, x_3)
                    th = np.acos(y/np.sqrt(x**2+y**2+z**2))
                    #print(coord, "=?sph2", th/np.pi)
                    # if abs(th) >= np.pi-self.theta_switch:# and abs(th) <= np.pi+self.theta_switch:
                    #     return 0
                    # else:
                    #     return 1
                
                if abs(th) <= self.theta_switch:# and abs(th) >= 2*np.pi-self.theta_switch:
                    return 0
                elif abs(th) >= np.pi-self.theta_switch:# and abs(th) <= np.pi+self.theta_switch:
                    return 0
                else:
                    return 1

                #return th - self.theta_switch

        result = self.integrator.integrate(\
                        k_0, x_0, self.metric.get_dk, hit_blackhole, \
                        stop_integration = stop_integration,\
                        stop_integration_coord_check = switch_coordinates,\
                        curve_start = curve_start, curve_end = curve_end, nr_points_curve = nr_points_curve, \
                        method = method, \
                        max_step = max_step, first_step = first_step, \
                        rtol = rtol, atol = atol,\
                        verbose = verbose )
                       

        k_t, k_r, k_th, k_ph, t, r, th, ph = result.y
        lamb = result.t

        k_sph = np.array([k_r, k_th, k_ph])
        x_sph = np.array([r, th, ph])

        k4_sph = np.array([k_t, k_r, k_th, k_ph])
        x4_sph = np.array([t, r, th, ph])

        result.update({"k4_sph": k4_sph, "x4_sph": x4_sph})

        return k_sph, x_sph, result




import sympy as sp
import numpy as np
#from scipy.integrate import solve_ivp
from scipy.optimize import fsolve
import time
from curvedpy.utils.conversions import Conversions
from curvedpy.geodesics.blackhole_integrators.integrator_4D import Integrator4D
from curvedpy.metrics.schwarzschild_metric import SchwarzschildMetricSpherical

# -------------------------------
# ----NAMING CONVENTIONS USED----
# -------------------------------
#
# UPPER & LOWER INDICES
# _ indicates an upper/contravariant index
# __ indicates a lower/covariant index
# For example g__mu__nu is the metric g with two lower indices while g_mu_nu is the metric 
# with two upper indices, which is the inverse of g__mu__nu
# Another example is 4-momentum k_mu, which has a 0 (time) component of k_0 or k_t
# The 4-momentum as a oneform is then k__mu and its zeroth component is k__0
# An example tensor could be T_mu__nu__alp. This tensor has one upper/contravariant (mu) and 
# two lower/covariant/oneform (nu, alp) indices.
#
# COORDINATES
# the used coordinate system for a tensor is indicated with by appending for example 
# _xyz, _sph, _bl after the indices.
# Example: x_mu_bl: this 4vector with one upper index is given in Boyer-Lindquist coordinates
# Example: g__mu__nu_sph: this metric tensor with two lower indices is given in spherical coordinates
#
# MISC
# If a vector x_mu has only 3 components, they are the three spatial components


# https://f.yukterez.net/einstein.equations/files/8.html#transformation
# https://physics.stackexchange.com/questions/672252/how-to-compute-and-interpret-schwartzchild-black-hole-metric-line-element-ds-i
class GeodesicIntegratorSchwarzschildSPH:

    conversions = Conversions()
    integrator = Integrator4D()

    ################################################################################################
    #
    ################################################################################################
    def __init__(self, mass=1.0, time_like = False, eps_theta=0.0000001, verbose=False):
        self.metric = SchwarzschildMetricSpherical(mass=mass, eps_theta=eps_theta)

    ################################################################################################
    #
    ################################################################################################
    def calc_trajectory(self, k0_xyz, x0_xyz, *args, **kargs):
        #mp_on = False

        if not isinstance(k0_xyz, np.ndarray): k0_xyz = np.array(k0_xyz)
        if not isinstance(x0_xyz, np.ndarray): x0_xyz = np.array(x0_xyz)


        if k0_xyz.shape != x0_xyz.shape:
            print("k and x are not the same shape")
            return

        if k0_xyz.ndim == 1:
            if k0_xyz.shape[0] != 3 or x0_xyz.shape[0] != 3:
                print("k or x do not have 3 components")
                return
            k0_xyz = k0_xyz.reshape(1,3)
            x0_xyz = x0_xyz.reshape(1,3)

        else:
            if k0_xyz.shape[1] != 3 or x0_xyz.shape[1] != 3:
                print("k or x do not have 3 components")
                return

        if len(k0_xyz) == 1:
            return self.calc_trajectory_xyz(k0_xyz[0], x0_xyz[0], *args, **kargs)
        else:
            return [self.calc_trajectory_xyz(k0_xyz[i], x0_xyz[i], *args, **kargs) for i in range(len(x0_xyz))]


    ################################################################################################
    #
    ################################################################################################
    def calc_trajectory_xyz(self, \
                        k0_xyz = np.array([1, 0.0, 0.0]), x0_xyz = np.array([-10, 10, 0]), \
                        R_end = -1,\
                        curve_start = 0, \
                        curve_end = 50, \
                        nr_points_curve = 50, \
                        method = "RK45",\
                        max_step = np.inf,\
                        first_step = None,\
                        rtol = 1e-3,\
                        atol = 1e-6,\
                        verbose = False \
                       ):

        if not isinstance(x0_xyz,np.ndarray):
            x0_xyz = np.array(x0_xyz)

        if not isinstance(k0_xyz,np.ndarray):
            k0_xyz = np.array(k0_xyz)

        x0_sph, k0_sph = self.conversions.convert_xyz_to_sph(x0_xyz, k0_xyz)
        k_r_0, k_th_0, k_ph_0 = k0_sph
        r0, th0, ph0 = x0_sph
        # Calculate from norm of starting condition
        t0 = 0
        k_t_0 = self.metric.k_t_from_norm_lamb(k_r_0, k_th_0, k_ph_0, t0, r0, th0, ph0)

        k_0 = np.array([k_t_0, *k0_sph])
        x_0 = np.array([t0, *x0_sph])

        #Check if starting values are outside the blackhole
        if r0 < self.metric.get_r_s():
            print("Starting value inside blackhole")
            return

        def hit_blackhole(t, y): 
            eps = 0.01
            k_0, k_1, k_2, k_3, x_0, x_1, x_2, x_3 = y
            r = x_1#calc_radius_from_x_mu(x_0, x_1, x_2, x_3)
            return r - (self.metric.get_r_s()+eps)

        if R_end == -1: R_end = np.inf
        #elif R_end < r0: R_end = r0*1.01
        def stop_integration(t, y): 
            k_0, k_1, k_2, k_3, x_0, x_1, x_2, x_3 = y
            r = x_1 #calc_radius_from_x_mu(x_0, x_1, x_2, x_3)
            return r - R_end

        result = self.integrator.integrate(\
                        k_0, x_0, self.metric.get_dk, hit_blackhole, \
                        stop_integration = stop_integration,\
                        curve_start = curve_start, curve_end = curve_end, nr_points_curve = nr_points_curve, \
                        method = method, \
                        max_step = max_step, first_step = first_step, \
                        rtol = rtol, atol = atol,\
                        verbose = verbose )
                       

        k_t, k_r, k_th, k_ph, t, r, th, ph = result.y
        lamb = result.t

        k_sph = np.array([k_r, k_th, k_ph])
        x_sph = np.array([r, th, ph])

        k4_sph = np.array([k_t, k_r, k_th, k_ph])
        x4_sph = np.array([t, r, th, ph])

        result.update({"k4_sph": k4_sph, "x4_sph": x4_sph})

        # SHOULD I NOT CHANGE COORDS USING 4 VECTORS????
        x_xyz, k_xyz = self.conversions.convert_sph_to_xyz(x_sph, k_sph)

        x4_xyz = np.array([t, *x_xyz])
        k4_xyz = np.array([k_t, *k_xyz])

        result.update({"k4_xyz": k4_xyz, "x4_xyz": x4_xyz})

        return k_xyz, x_xyz, result




import unittest
import curvedpy as cp
import sympy as sp
import numpy as np
import random
from curvedpy.utils.conversions import Conversions

################################################################################################
#
################################################################################################
class TestConversions(unittest.TestCase):
    def setUp(self):
        self.converter = Conversions()

    # A single test to see if the spherical and cartesian conversion works
    def test_CONV_sph_to_xyz_and_back(self):
        k0_xyz = np.array([11.322145, 15.136237, 65.246265])
        x0_xyz = np.array([13.461341, 13.461346, 72.300564])
        round_lvl = 6

        x0_sph, k0_sph = self.converter.convert_xyz_to_sph(x0_xyz, k0_xyz)
        x0_xyz_new, k0_xyz_new = self.converter.convert_sph_to_xyz(x0_sph, k0_sph)

        self.assertTrue( bool((k0_xyz == [round(v, round_lvl) for v in k0_xyz_new]).all()) )
        self.assertTrue( bool((x0_xyz == [round(v, round_lvl) for v in x0_xyz_new]).all()) )

    # Sample of tests to see if the conversion from Boyer-Lindquist to xyz and back conversion works
    def test_CONV_xyz_to_bl_and_back_conversions(self):
        round_lvl = 13
        xl, yl, zl = [], [], []
        kxl, kyl, kzl = [], [], []
        for i in range(100):
            x0_xyz = np.array([random.uniform(-100, 100), random.uniform(-100, 100), random.uniform(-100, 100)])
            k0_xyz = np.array([random.uniform(-100, 100), random.uniform(-100, 100), random.uniform(-100, 100)])
            
            x0, y0, z0 = x0_xyz
            kx0, ky0, kz0 = k0_xyz
            
            x_bl, k_bl = self.converter.convert_xyz_to_bl(x0_xyz, k0_xyz, a=1)
            
            x_xyz, k_xyz = self.converter.convert_bl_to_xyz(x_bl, k_bl, a=1)

            x, y, z = x_xyz
            kx, ky, kz = k_xyz
            
            xl.append(x-x0)
            yl.append(y-y0)
            zl.append(z-z0)
            kxl.append(kx-kx0)
            kyl.append(ky-ky0)
            kzl.append(kz-kz0)

        np.std(xl), np.std(yl), np.std(zl), np.std(kxl), np.std(kyl), np.std(kzl)
        self.assertTrue( round(np.std(xl), round_lvl) == 0.0)

if __name__ == '__main__':
    unittest.main()


import sympy as sp
import numpy as np

class Conversions:
    def __init__(self, verbose=False):

        # NOTE TO SELF: I treat things as 3 vectors here. What if the metric has a mixed term, should I not include t? (26/11/2024)
        self.setup_sph_to_xyz_conversion()
        self.setup_xyz_to_sph_conversion()
        self.setup_bl_to_xyz_conversion()
        self.setup_xyz_to_bl_conversion()

    def setup_xyz_to_bl_conversion(self):
        x, y, z, a = sp.symbols(" x y z a")
        k_x, k_y, k_z = sp.symbols(" k_x k_y k_z")

        r = (1/sp.sqrt(2)) * sp.sqrt(sp.sqrt(z**4 + 2 * y**2 * z**2 + 2*x**2 * z**2 + 2*a**2 * z**2 + y**4 + 2*x**2*y**2 - 2*a**2*y**2 + x**4 - 2*a**2*x**2 + a**4) +x**2+y**2+z**2-a**2)
        th = sp.acos(z/r)
        phi = sp.atan2(y,x)

        M_xyz_to_bl = sp.Matrix([[r.diff(x), r.diff(y), r.diff(z)],\
                                  [th.diff(x), th.diff(y), th.diff(z)],\
                                  [phi.diff(x), phi.diff(y), phi.diff(z)],\
                                 ])
        
        k = sp.Matrix([k_x, k_y, k_z])
        k_bl = M_xyz_to_bl*k
        self.convert_k_xyz_to_k_bl = sp.lambdify([x, y, z, k_x, k_y, k_z, a], \
                                     k_bl, "numpy")

    def setup_bl_to_xyz_conversion(self):
        r, th, ph, a = sp.symbols(" r th ph a")
        k_r, k_th, k_ph = sp.symbols(" k_r k_th k_ph")

        x = sp.sqrt(r**2 + a**2) * sp.sin(th) * sp.cos(ph)
        y = sp.sqrt(r**2 + a**2) * sp.sin(th) * sp.sin(ph)
        z = r * sp.cos(th)
    
        M_bl_to_xyz = sp.Matrix([[x.diff(r), x.diff(th), x.diff(ph)],\
                                  [y.diff(r), y.diff(th), y.diff(ph)],\
                                  [z.diff(r), z.diff(th), z.diff(ph)],\
                                 ])

        v = sp.Matrix([r, th , ph])
        k = sp.Matrix([k_r, k_th, k_ph])
        k_xyz = M_bl_to_xyz*k
        self.convert_k_bl_to_k_xyz = sp.lambdify([r, th, ph, k_r, k_th, k_ph, a], \
                                     k_xyz, "numpy")

    def setup_xyz_to_sph_conversion(self):
        x, y, z = sp.symbols(" x y z ")
        k_x, k_y, k_z = sp.symbols(" k_x k_y k_z")

        r = sp.sqrt(x**2+y**2+z**2)
        th = sp.acos(z/r)
        phi = sp.atan(y/x)

        M_xyz_to_sph = sp.Matrix([[r.diff(x), r.diff(y), r.diff(z)],\
                                  [th.diff(x), th.diff(y), th.diff(z)],\
                                  [phi.diff(x), phi.diff(y), phi.diff(z)],\
                                 ])

        k = sp.Matrix([k_x, k_y, k_z])
        k_sph = M_xyz_to_sph*k
        self.convert_k_xyz_to_k_sph = sp.lambdify([x, y, z, k_x, k_y, k_z], \
                                     k_sph, "numpy")

    def setup_sph_to_xyz_conversion(self):
        r, th, ph = sp.symbols(" r th ph ")
        k_r, k_th, k_ph = sp.symbols(" k_r k_th k_ph")

        x = r * sp.sin(th) * sp.cos(ph)
        y = r * sp.sin(th) * sp.sin(ph)
        z = r * sp.cos(th)

        M_sph_to_xyz = sp.Matrix([[x.diff(r), x.diff(th), x.diff(ph)],\
                                  [y.diff(r), y.diff(th), y.diff(ph)],\
                                  [z.diff(r), z.diff(th), z.diff(ph)],\
                                 ])

        v = sp.Matrix([r, th , ph])
        k = sp.Matrix([k_r, k_th, k_ph])
        k_xyz = M_sph_to_xyz*k
        self.convert_k_sph_to_k_xyz = sp.lambdify([r, th, ph, k_r, k_th, k_ph], \
                                     k_xyz, "numpy")

    ##################################################################
    # SPHERICAL - XYZ
    ##################################################################
    def convert_sph_to_xyz(self, x_sph, k_sph):
        v_xyz = self.coord_conversion_sph_to_xyz(*x_sph)
        k_xyz = self.convert_k_sph_to_k_xyz(*x_sph, *k_sph)
        k_xyz = k_xyz.reshape(*k_sph.shape)
        return v_xyz, k_xyz

    def coord_conversion_sph_to_xyz(self, r, th, ph):
        z = r*np.cos(th)
        x = r*np.sin(th)*np.cos(ph)
        y = r*np.sin(th)*np.sin(ph)
        return np.array([x, y, z])

    ##################################################################
    # XYZ - SPHERICAL
    ##################################################################
    def convert_xyz_to_sph(self, x_xyz, k_xyz):
        v_sph = self.coord_conversion_xyz_to_sph(*x_xyz)
        k_sph = self.convert_k_xyz_to_k_sph(*x_xyz, *k_xyz)
        k_sph = k_sph.reshape(*k_xyz.shape)
        return v_sph, k_sph

    def coord_conversion_xyz_to_sph(self, x, y, z):
        r = np.sqrt(x**2 + y**2 + z**2)
        th = np.acos(z/r)
        ph = np.atan2(y, x) #ph = np.atan(y/x)
        return r, th, ph

    ##################################################################
    # BOYER-LINDQUIST - XYZ
    ##################################################################
    def convert_bl_to_xyz(self, x_bl, k_bl, a):
        v_xyz = self.coord_conversion_bl_xyz(*x_bl, a)
        k_xyz = self.convert_k_bl_to_k_xyz(*x_bl, *k_bl, a)
        k_xyz = k_xyz.reshape(*k_bl.shape)
        return v_xyz, k_xyz

    def coord_conversion_bl_xyz(self, r, th, ph, a):
        x = np.sqrt(r**2 + a**2) * np.sin(th) * np.cos(ph)
        y = np.sqrt(r**2 + a**2) * np.sin(th) * np.sin(ph)
        z = r * np.cos(th)
        return x, y, z

    ##################################################################
    # XYZ - BOYER-LINDQUIST
    ##################################################################
    def convert_xyz_to_bl(self, x_xyz, k_xyz, a):
        v_bl = self.coord_conversion_xyz_bl(*x_xyz, a)
        k_bl = self.convert_k_xyz_to_k_bl(*x_xyz, *k_xyz, a)
        k_bl = k_bl.reshape(*k_xyz.shape)
        return v_bl, k_bl

    def coord_conversion_xyz_bl(self, x, y, z, a):
        # https://physics.stackexchange.com/questions/739653/how-does-one-covert-from-cartesian-coordinates-to-boyer-lindquist-coordinates
        # at a = 0, r reduces to spherical coordinate r**2 = x**2+y**2+z**2
        r = (1/np.sqrt(2)) * np.sqrt(np.sqrt(z**4 + 2 * y**2 * z**2 + 2*x**2 * z**2 + 2*a**2 * z**2 + y**4 + 2*x**2*y**2 - 2*a**2*y**2 + x**4 - 2*a**2*x**2 + a**4) +x**2+y**2+z**2-a**2)
        th = np.acos(z/r)
        ph = np.atan2(y,x)
        return r, th, ph

    ####################################################
    # Sympy based functions
    ####################################################
    #def vec_to_oneform(self, vec, metric):








import sympy as sp
import numpy as np
from scipy.integrate import solve_ivp
import time


class GeodesicIntegratorIsotropicXYZ:
    def __init__(self, metric = "schwarzschild_isotropic", mass = 1., verbose=False):

        # Schwarzschild radius together with mass in geometrized units
        self.M = mass
        self.r_s_value = 2*self.M 

        # Type of geodesic
        self.time_like = False # No Massive particle geodesics yet

        # Define symbolic variables
        self.t, self.x, self.y, self.z, self.r_s = sp.symbols('t x y z r_s')

        # Radial distance to BlackHole location
        self.R = sp.sqrt(self.x**2 + self.y**2 + self.z**2)

        # The implemented metrics:
        if metric == "flat":
            self.g = sp.Matrix([\
                [-1, 0, 0, 0],\
                [0, 1, 0, 0],\
                [0, 0, 1, 0],\
                [0, 0, 0, 1]\
                ])
        elif metric == "schwarzschild_isotropic":
            self.g = sp.Matrix([\
                [-(1-self.r_s/(4*self.R))**2 / (1+self.r_s/(4*self.R))**2, 0, 0, 0],\
                [0, (1+self.r_s/(4*self.R))**4, 0, 0], \
                [0, 0, (1+self.r_s/(4*self.R))**4, 0], \
                [0, 0, 0, (1+self.r_s/(4*self.R))**4], \
              ]) 
        
        # Connection Symbols
        self.gam_t = sp.Matrix([[self.gamma_func(0,mu, nu).simplify() for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        self.gam_x = sp.Matrix([[self.gamma_func(1,mu, nu).simplify() for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        self.gam_y = sp.Matrix([[self.gamma_func(2,mu, nu).simplify() for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        self.gam_z = sp.Matrix([[self.gamma_func(3,mu, nu).simplify() for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        if verbose: print("Done connection symbols")
        
        # Building up the geodesic equation: 
        # Derivatives: k_beta = d x^beta / d lambda
        self.k_t, self.k_x, self.k_y, self.k_z = sp.symbols('k_t k_x k_y k_z', real=True)
        self.k = [self.k_t, self.k_x, self.k_y, self.k_z]
    
        # Second derivatives: d k_beta = d^2 x^beta / d lambda^2
        self.dk_t = sum([- self.gam_t[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        self.dk_x = sum([- self.gam_x[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        self.dk_y = sum([- self.gam_y[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        self.dk_z = sum([- self.gam_z[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        print(self.dk_z)
        if verbose: print("Done diff of k")

        # Norm of k
        # the norm of k determines if you have a massive particle (-1), a mass-less photon (0) 
        # or a space-like curve (1)
        self.k = sp.Matrix([self.k_t, self.k_x, self.k_y, self.k_z])
        self.norm_k = (self.k.T*self.g*self.k)[0]
        self.norm_k_lamb = sp.lambdify([self.k_t, self.k_x, self.x, self.k_y, self.y, self.k_z, self.z, \
                                               self.r_s], self.norm_k, "numpy")

        # Now we calculate k_t using the norm. This eliminates one of the differential equations.
        # time_like = True: calculates a geodesic for a massive particle (not implemented yet)
        # time_like = False: calculates a geodesic for a photon
        if (self.time_like):
            self.k_t_from_norm = sp.solve(self.norm_k+1, self.k_t)[1]
        else:
            self.k_t_from_norm = sp.solve(self.norm_k, self.k_t)[1]
        if verbose: print("Done norm of k")

        # Lambdify versions
        self.dk_x_lamb = sp.lambdify([self.k_x, self.x, self.k_y, self.y, self.k_z, self.z, \
                                      self.k_t, self.t, self.r_s], \
                                     self.dk_x, "numpy")
        self.dk_y_lamb = sp.lambdify([self.k_x, self.x, self.k_y, self.y, self.k_z, self.z, \
                                      self.k_t, self.t, self.r_s], \
                                     self.dk_y, "numpy")
        self.dk_z_lamb = sp.lambdify([self.k_x, self.x, self.k_y, self.y, self.k_z, self.z, \
                                      self.k_t, self.t, self.r_s], \
                                     self.dk_z, "numpy")
        self.k_t_from_norm_lamb = sp.lambdify([self.k_x, self.x, self.k_y, self.y, self.k_z, self.z, \
                                               self.r_s], self.k_t_from_norm, "numpy")
        if verbose: print("Done lambdifying")
     
    # Connection Symbols
    def gamma_func(self, sigma, mu, nu):
        coord_symbols = [self.t, self.x, self.y, self.z]
        g_sigma_mu_nu = 0
        for rho in [0,1,2,3]:
            if self.g[sigma, rho] != 0:
                g_sigma_mu_nu += 1/2 * 1/self.g[sigma, rho] * (\
                                self.g[nu, rho].diff(coord_symbols[mu]) + \
                                self.g[rho, mu].diff(coord_symbols[nu]) - \
                                self.g[mu, nu].diff(coord_symbols[rho]) )
            else:
                g_sigma_mu_nu += 0
        return g_sigma_mu_nu

    # This function does the numerical integration of the geodesic equation using scipy's solve_ivp
    def calc_trajectory(self, \
                        k_x_0 = 1., x0 = -10.0, k_y_0 = 0., y0 = 5.0, k_z_0 = 0., z0 = 5.0,\
                        R_end = -1,\
                        curve_start = 0, \
                        curve_end = 50, \
                        nr_points_curve = 50, \
                        method = "RK45",\
                        max_step = np.inf,\
                        first_step = None,\
                        rtol = 1e-3,\
                        atol = 1e-6,\
                        verbose = False \
                       ):

        r0 = np.linalg.norm(np.array([x0, y0, z0]))
        if r0 > self.r_s_value:
            # Step function needed for solve_ivp
            def step(lamb, new):

                new_k_x, new_x, new_k_y, new_y, new_k_z, new_z = new
                new_k_t = self.k_t_from_norm_lamb(*new, self.r_s_value)

                #print(np.sqrt(new_x**2 + new_y**2 + new_z**2))

                new_dk_x = self.dk_x_lamb(*new, new_k_t, t = 0, r_s = self.r_s_value)
                dx = new_k_x
                new_dk_y = self.dk_y_lamb(*new, new_k_t, t = 0, r_s = self.r_s_value)
                dy = new_k_y
                new_dk_z = self.dk_z_lamb(*new, new_k_t, t = 0, r_s = self.r_s_value)
                dz = new_k_z
                #print("step", new_x, new_y, new_z, dx, dy, dz)

                return( new_dk_x, dx, new_dk_y, dy, new_dk_z, dz)

            def hit_blackhole(t, y): 
                eps = 0.5
                k_x, x, k_y, y, k_z, z = y
                if verbose: print("Test Event Hit BH: ", x, y, z, self.r_s_value, x**2 + y**2 + z**2 - self.r_s_value**2)
                return x**2 + y**2 + z**2 - (self.r_s_value+eps)**2
            hit_blackhole.terminal = True

            def reached_end(t, y): 
                k_x, x, k_y, y, k_z, z = y
                if verbose: print("Test Event End: ", np.sqrt(x**2 + y**2 + z**2), R_end, x**2 + y**2 + z**2 - R_end**2)
                return x**2 + y**2 + z**2 - R_end**2
            reached_end.terminal = True
            
            values_0 = [ k_x_0, x0, k_y_0, y0, k_z_0, z0 ]
            if nr_points_curve == 0:
                t_pts = None
            else:
                t_pts = np.linspace(curve_start, curve_end, nr_points_curve)

            start = time.time()
            events = [hit_blackhole]
            if R_end > r0 : events.append(reached_end)
            result = solve_ivp(step, (curve_start, curve_end), values_0, t_eval=t_pts, \
                               events=events,\
                               method=method,\
                               max_step = max_step,\
                               first_step = first_step,\
                               atol=atol,\
                               rtol = rtol)
            end = time.time()
            if verbose: print("New: ", result.message, end-start, "sec")


            result.update({"hit_blackhole": len(result.t_events[0])>0})
            result.update({"start_inside_hole": False})

        else:
            if verbose: print("Starting location inside the blackhole.")
            result = {"start_inside_hole": True}

        return result



import sympy as sp
import numpy as np
from scipy.integrate import solve_ivp
import time
#import multiprocessing as mp
from curvedpy.utils.conversions import Conversions

class GeodesicIntegratorKerr:

    conversions = Conversions()

    def __init__(self, \
            mass=1.0, a = 0.0, time_like = False, \
            pre_sub_metric_params = False, \
            calc_kt_per_ray=True, \
            simplify_inv=True, \
            verbose=False, verbose_init = False ):

        if verbose_init: print("Integrator - started init")

        self.calc_kt_per_ray = calc_kt_per_ray

        # Connection Symbols
        def gamma_func(sigma, mu, nu):
            coord_symbols = [self.t, self.r, self.th, self.ph]
            g_sigma_mu_nu = 0
            for rho in [0,1,2,3]:
                if self.g[sigma, rho] != 0:
                    g_sigma_mu_nu += 1/2 * self.g_inv[sigma, rho] * (\
                                    self.g_diff[mu][nu, rho] + \
                                    self.g_diff[nu][rho, mu] - \
                                    self.g_diff[rho][mu, nu] )
                else:
                    g_sigma_mu_nu += 0
            return g_sigma_mu_nu

        self.M = mass
        self.r_s_value = 2*self.M 
        self.a_value = a # Remember that most of the time we integrate backwards in time and the rotation of the blackhole is reversed!

        # Type of geodesic
        self.time_like = time_like

        self.verbose = verbose 


        if verbose_init:
            print("Geodesic Kerr Integrator Settings: ")
            print(f"  - {self.M=}")
            print(f"  - {self.a_value=}")
            print(f"  - {self.r_s_value=}")
            print(f"  - {self.time_like=}")
            print(f"  - {self.verbose=} does not work from init yet")
            print(f"  - {pre_sub_metric_params=}")
            print(f"  - {self.calc_kt_per_ray=}")
            print(f"  - {simplify_inv=}")
            print("--")

        # Define symbolic variables
        self.t, self.r, self.th, self.ph, self.r_s, self.a = sp.symbols("t r \\theta \\phi r_s a", positive=True, real=True)
        self.Sig, self.Del = sp.symbols("\\Sigma \\Delta")
        self.Sig_sub = self.r**2 + self.a**2 * sp.cos(self.th)**2
        self.Del_sub = self.r**2 - self.r_s*self.r + self.a**2

        g00 = -(1-self.r_s*self.r/self.Sig)
        g11 = self.Sig/self.Del
        g22 = self.Sig
        #g33 = (self.r**2 + self.a**2 + self.r_s*self.r*self.a**2 * sp.sin(self.th)**2 / self.Sig )*sp.sin(self.th)**2 
        g33 = (sp.sin(self.th)**2/self.Sig)*((self.r**2 + self.a**2)**2 - self.a**2 * self.Del *sp.sin(self.th)**2)
        g03 = -(self.r_s * self.r * self.a * sp.sin(self.th)**2/self.Sig) # Carefull, no 2* here since it comes into the metric twice at 03 and 30

        # The metrix in terms of Sigma and Delta
        self.g_simple = sp.Matrix([\
            [g00, 0, 0, g03],\
            [0, g11, 0, 0],\
            [0, 0, g22, 0],\
            [g03, 0, 0, g33]\
            ])

        # For the Christoffel symbols we need the inverse of the metric
        # For the Kerr metric, which is not diagonal, we need to do this
        # properly
        if verbose_init: start = time.time()
        self.g_simple_inv = self.g_simple.inv()
        if simplify_inv: self.g_simple_inv.simplify()
        if verbose_init: print("Integrator - done inverse g in (sec): ", round(time.time()-start, 5))

        # Defining the four momentum per mass to use in the Geodesic Equation
        # Derivatives: k_beta = d x^beta / d lambda
        self.k_t, self.k_r, self.k_th, self.k_ph = sp.symbols('k_t k_r k_th k_ph', real=True)
        self.k = sp.Matrix([self.k_t, self.k_r, self.k_th, self.k_ph])
        # Also calculate the 1 form of the momentum per mass 4 vector
        self.k__ = self.g_simple*self.k
        self.k__ = self.k__.subs([(self.Del, self.Del_sub), (self.Sig, self.Sig_sub)])

        # Norm of k
        # the norm of k determines if you have a massive particle (-1), a mass-less photon (0) 
        # or a space-like curve (1)
        self.norm_k = (self.k.T*self.g_simple*self.k)[0]

        # Now we calculate k_t using the norm. This eliminates one of the differential equations.
        # time_like = True: calculates a geodesic for a massive particle
        # time_like = False: calculates a geodesic for a photon
        if verbose_init: start = time.time()
        if not self.calc_kt_per_ray:
            if (self.time_like):
                self.k_t_from_norm = sp.solve(self.norm_k+1, self.k_t)[1]
            else:
                self.k_t_from_norm = sp.solve(self.norm_k, self.k_t)[1]
        if verbose_init: print("Integrator - Done norm of k in:", round(time.time()-start, 5))
        

        # Getting the metric in full and only dependend on the coordinates
        self.g = self.g_simple.subs([(self.Del, self.Del_sub), (self.Sig, self.Sig_sub)])
        #self.g = self.g.subs(self.Sig, self.Sig_sub)
        self.g_inv = self.g_simple_inv.subs([(self.Del, self.Del_sub), (self.Sig, self.Sig_sub)])
        self.norm_k = self.norm_k.subs([(self.Del, self.Del_sub), (self.Sig, self.Sig_sub)])
        
        if pre_sub_metric_params:
            # Subbing these in can make things a lot faster in some cases.
            self.g = self.g.subs([(self.a, self.a_value), (self.r_s, self.r_s_value)])
            self.g_inv = self.g_inv.subs([(self.a, self.a_value), (self.r_s, self.r_s_value)])

        # We already differentiate the metric to all the coordinates, otherwise we
        # will do this multiple time for the same elements when we calculating the 
        # Christoffel symbols
        if verbose_init: start = time.time()
        self.g_diff = [self.g.diff(self.t), self.g.diff(self.r), self.g.diff(self.th), self.g.diff(self.ph)]
        if verbose_init: print("Integrator - done with g_diff in: ", round(time.time()-start, 5))


        # Time to start calculating the Christoffel Symbols
        if verbose_init: start = time.time()
        self.gam_t = sp.Matrix([[gamma_func(0,mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        self.gam_r = sp.Matrix([[gamma_func(1,mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        self.gam_th = sp.Matrix([[gamma_func(2,mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        self.gam_ph = sp.Matrix([[gamma_func(3,mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        if verbose_init: print("Integrator - done connection symbols in: ", round(time.time()-start, 5))

        #self.k = [self.k_t, self.k_r, self.k_th, self.k_ph]
        

        # Calculating the directional derivative of the four momentum per mass, also to use in the Geodesic equation
        # Second derivatives: d k_beta = d^2 x^beta / d lambda^2
        if verbose_init: start = time.time()
        self.dk_t = sum([- self.gam_t[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        self.dk_r = sum([- self.gam_r[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        self.dk_th = sum([- self.gam_th[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        self.dk_ph = sum([- self.gam_ph[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        if verbose_init: print("Integrator - Done diff of k in: ", round(time.time()-start, 5))


        # Lambdify versions of the directional derivatives of the four momentum per mass
        if verbose_init: start = time.time()
        self.dk_t_lamb = sp.lambdify([self.k_r, self.r, self.k_th, self.th, self.k_ph, self.ph, \
                                      self.k_t, self.t, self.r_s, self.a], \
                                     self.dk_t, "numpy")
        self.dk_r_lamb = sp.lambdify([self.k_r, self.r, self.k_th, self.th, self.k_ph, self.ph, \
                                      self.k_t, self.t, self.r_s, self.a], \
                                     self.dk_r, "numpy")
        self.dk_th_lamb = sp.lambdify([self.k_r, self.r, self.k_th, self.th, self.k_ph, self.ph, \
                                      self.k_t, self.t, self.r_s, self.a], \
                                     self.dk_th, "numpy")
        self.dk_ph_lamb = sp.lambdify([self.k_r, self.r, self.k_th, self.th, self.k_ph, self.ph, \
                                      self.k_t, self.t, self.r_s, self.a], \
                                     self.dk_ph, "numpy")

        self.norm_k_lamb = sp.lambdify([self.k_t, self.k_r, self.r, self.k_th, self.th, self.k_ph, self.ph, \
                                               self.r_s, self.a], self.norm_k, "numpy")

        if not self.calc_kt_per_ray:
            self.k_t_from_norm = self.k_t_from_norm.subs([(self.Del, self.Del_sub), (self.Sig, self.Sig_sub)])
            self.k_t_from_norm_lamb = sp.lambdify([self.k_r, self.r, self.k_th, self.th, self.k_ph, self.ph, \
                                                   self.r_s, self.a], self.k_t_from_norm, "numpy")

        self.lamb_k__ = sp.lambdify([self.k_t, self.k_r, self.k_th, self.k_ph, self.t, self.r, self.th, self.ph, self.r_s, self.a],\
                                 self.k__, "numpy")

        if verbose_init: 
            print("Inegrator - Done lambdifying in:", round(time.time()-start, 5))
            print("Integrator - init done")
            print()




    ################################################################################################
    #
    ################################################################################################
    def get_k_t_from_norm(self, k_r_0, r0, k_th_0, th0, k_ph_0, ph0):
        sub_list = [(self.t, 0), (self.r, r0), (self.th, th0), (self.ph, ph0), \
                    (self.k_r, k_r_0), (self.k_th, k_th_0), (self.k_ph, k_ph_0), \
                    (self.r_s, self.r_s_value), (self.a, self.a_value)]

        if (self.time_like):
            k_t_from_norm = sp.solve(self.norm_k.subs(sub_list)+1, self.k_t)[1]#[1] #sp.solve(self.norm_k+1, self.k_t)[1]
        else:
            k_t_from_norm = sp.solve(self.norm_k.subs(sub_list), self.k_t)[1]#[1] #sp.solve(self.norm_k, self.k_t)[1]

        
        return k_t_from_norm

    ################################################################################################
    #
    ################################################################################################
    def calc_trajectory(self, k0_xyz, x0_xyz, *args, **kargs):
        #mp_on = False

        if not isinstance(k0_xyz, np.ndarray): k0_xyz = np.array(k0_xyz)
        if not isinstance(x0_xyz, np.ndarray): x0_xyz = np.array(x0_xyz)

        if k0_xyz.shape != x0_xyz.shape:
            print("k and x are not the same shape")
            return

        if k0_xyz.ndim == 1:
            if k0_xyz.shape[0] != 3 or x0_xyz.shape[0] != 3:
                print("k or x do not have 3 components")
                return
            k0_xyz = k0_xyz.reshape(1,3)
            x0_xyz = x0_xyz.reshape(1,3)

        else:
            if k0_xyz.shape[1] != 3 or x0_xyz.shape[1] != 3:
                print("k or x do not have 3 components")
                return

        if len(k0_xyz) == 1:
            return self.calc_trajectory_xyz(k0_xyz[0], x0_xyz[0], *args, **kargs)
        else:
            return [self.calc_trajectory_xyz(k0_xyz[i], x0_xyz[i], *args, **kargs) for i in range(len(x0_xyz))]


    ################################################################################################
    #
    ################################################################################################
    def calc_trajectory_xyz(self, \
                        k0_xyz = np.array([1, 0.0, 0.0]), x0_xyz = np.array([-10, 10, 0]), \
                        R_end = -1,\
                        curve_start = 0, \
                        curve_end = 50, \
                        nr_points_curve = 50, \
                        method = "RK45",\
                        max_step = np.inf,\
                        first_step = None,\
                        rtol = 1e-3,\
                        atol = 1e-6,\
                        verbose = False \
                       ):

        if verbose: print("Calculate Traj Started")
        if verbose: print(f"{max_step=}")
        if not isinstance(x0_xyz,np.ndarray):
            x0_xyz = np.array(x0_xyz)

        if not isinstance(k0_xyz,np.ndarray):
            k0_xyz = np.array(k0_xyz)

        x0_bl, k0_bl = self.conversions.convert_xyz_to_bl(x0_xyz, k0_xyz, a = self.a_value)
        k_r_0, k_th_0, k_ph_0 = k0_bl
        r0, th0, ph0 = x0_bl

        result = self.calc_trajectory_bl(\
                        k_r_0 = k_r_0, r0 = r0, k_th_0=k_th_0, th0=th0, k_ph_0=k_ph_0, ph0=ph0,\
                        R_end = R_end, curve_start = curve_start, curve_end = curve_end, nr_points_curve = nr_points_curve, \
                        method = method, max_step = max_step, first_step = first_step, rtol = rtol, atol = atol,\
                        verbose = verbose )
                       

        k_r, r, k_th, th, k_ph, ph, k_t = result.y
        t = result.t

        k_bl = np.array([k_r, k_th, k_ph])
        x_bl = np.array([r, th, ph])

        if verbose: start = time.time()
        # SHOULD I NOT CHANGE COORDS USING 4 VECTORS????
        x_xyz, k_xyz = self.conversions.convert_bl_to_xyz(x_bl, k_bl, self.a_value)
        if verbose: print("  Converting result to xyz in: ", round(time.time()-start, 5))

        x4_xyz = np.array([t, *x_xyz])
        k4_xyz = np.array([k_t, *k_xyz])

        result.update({"k4_xyz": k4_xyz, "x4_xyz": x4_xyz})

        return k_xyz, x_xyz, result

    ################################################################################################
    # calc_trajectory_bl
    ################################################################################################
    # This function does the numerical integration of the geodesic equation using scipy's solve_ivp
    def calc_trajectory_bl(self, \
                        k_r_0 = 0., r0 = 10.0, k_th_0 = 0.0, th0 = 1/2*np.pi, k_ph_0 = 0.1, ph0 = 0.0,\
                        R_end = -1,\
                        curve_start = 0, \
                        curve_end = 50, \
                        nr_points_curve = 50, \
                        method = "RK45",\
                        max_step = np.inf,\
                        first_step = None,\
                        rtol = 1e-3,\
                        atol = 1e-6,\
                        verbose = False \
                       ):

        # Calculate from norm of starting condition
        if verbose: start = time.time()
        if self.calc_kt_per_ray:
            k_t_0 = self.get_k_t_from_norm(k_r_0, r0, k_th_0, th0, k_ph_0, ph0)
        else:
            k_t_0 = self.k_t_from_norm_lamb(k_r_0, r0, k_th_0, th0, k_ph_0, ph0, self.r_s_value, self.a_value)
        if verbose: print("  Calculated k_t in: ", round(time.time()-start, 5))

        if R_end == -1:
            R_end = np.inf
        elif R_end < r0:
            R_end = r0*1.01

        if r0 > self.r_s_value:
            # Step function needed for solve_ivp
            def step(lamb, new):

                # if verbose:
                #     k_r, r, k_th, th, k_ph, ph, k_t = new
                #     x, y, z = self.conversions.coord_conversion_bl_xyz(r, th, ph, self.a_value)
                #     print(round(x, 4), round(y,4), round(z,4))

                new_k_r, new_r, new_k_th, new_th, new_k_ph, new_ph, new_k_t = new

                new_dk_t = self.dk_t_lamb(*new, t = lamb, r_s = self.r_s_value, a = self.a_value)
                dr = new_k_r
                new_dk_r = self.dk_r_lamb(*new, t = lamb, r_s = self.r_s_value, a = self.a_value)
                dr = new_k_r
                new_dk_th = self.dk_th_lamb(*new, t = lamb, r_s = self.r_s_value, a = self.a_value)
                dth = new_k_th
                new_dk_ph = self.dk_ph_lamb(*new, t = lamb, r_s = self.r_s_value, a = self.a_value)
                dph = new_k_ph

                return( new_dk_r, dr, new_dk_th, dth, new_dk_ph, dph, new_dk_t)

            def hit_blackhole(t, y): 
                eps = 0.01
                k_r, r, k_th, th, k_ph, ph, k_t = y
                #if verbose: print("Event - hit_blackhole: ", r-self.r_s_value)
                return r - (self.r_s_value+eps)
            hit_blackhole.terminal = True

            def reached_end(t, y): 
                k_r, r, k_th, th, k_ph, ph, k_t = y
                return r - R_end
            reached_end.terminal = True
            
            values_0 = [ k_r_0, r0, k_th_0, th0, k_ph_0, ph0, k_t_0 ]
            if nr_points_curve == 0:
                t_pts = None
            else:
                t_pts = np.linspace(curve_start, curve_end, nr_points_curve)

            events = [hit_blackhole]
            events.append(reached_end)

            start = time.time()
            result = solve_ivp(step, (curve_start, curve_end), values_0, t_eval=t_pts, \
                               events=events,\
                               method=method,\
                               max_step = max_step,\
                               first_step = first_step,\
                               atol=atol,\
                               rtol = rtol)
            if verbose: 
                print("  Finished solver: ", result.message)
                print("   Blackhole hit? ", len(result.t_events[0])>0)
                print("   Time: ",round(time.time()-start, 5), "sec")

            result.update({"R_end": R_end})
            result.update({"hit_blackhole": len(result.t_events[0])>0})
            result.update({"start_inside_hole": False})
            result.update({"end_check": len(result.t_events[1])>0})


        else:
            if verbose: print("Starting location inside the blackhole.")
            result = {"start_inside_hole": True}

        return result


    ################################################################################################
    # Helper functions
    ################################################################################################

    def one_form(self, vec4, x4):
        return self.one_form_lamb(*vec4, *x4)
        # #tv, xv, yv, zv = x
        # #rv, thv, phv = self.conversions.coord_conversion_xyz_bl(xv, yv, zv, self.a_value)
        # tv, rv, thv, phv = x_bl
        # sub_list = [(self.t, tv), (self.r, rv), (self.th, thv), (self.ph, phv), (self.r_s, self.r_s_value), (self.a, self.a_value)]
        # #print(sub_list)
        # return self.g.subs(sub_list)

    # Conserved angular momentum per mass
    def ang_mom(self, r, k_ph):
        return r**2*k_ph

    def energy_photon(self, k_r, r, k_ph, M_blackhole):
        return np.sqrt(k_r**2 +(1-2*M_blackhole/r)*self.ang_mom(r, k_ph)**2/r**2)

    def energy_massive(self, k_r, r, k_ph, M_blackhole):
        return np.sqrt(k_r**2 +(1-2*M_blackhole/r)*(1+self.ang_mom(r, k_ph)**2/r**2))



import sympy as sp
import numpy as np

class Conversions4D:
    def __init__(self, verbose=False):

        self.setup_sph_to_xyz_conversion()
        self.setup_xyz_to_sph_conversion()
        #self.setup_bl_to_xyz_conversion()
        #self.setup_xyz_to_bl_conversion()

    def setup_xyz_to_sph_conversion(self):
        t, x, y, z = sp.symbols("t x y z ")
        k_t, k_x, k_y, k_z = sp.symbols("k_t k_x k_y k_z")

        r = sp.sqrt(x**2+y**2+z**2)
        th = sp.acos(z/r)
        phi = sp.atan2(y,x)

        M_xyz_to_sph = sp.Matrix([[1, 0, 0, 0],\
                                  [0, r.diff(x), r.diff(y), r.diff(z)],\
                                  [0, th.diff(x), th.diff(y), th.diff(z)],\
                                  [0, phi.diff(x), phi.diff(y), phi.diff(z)],\
                                 ])

        k = sp.Matrix([k_t, k_x, k_y, k_z])
        k_sph = M_xyz_to_sph*k
        self.convert_k_xyz_to_k_sph = sp.lambdify([t, x, y, z, k_t, k_x, k_y, k_z], \
                                     k_sph, "numpy")

    def setup_sph_to_xyz_conversion(self):
        t, r, th, ph = sp.symbols("t r th ph ")
        k_t, k_r, k_th, k_ph = sp.symbols("k_t k_r k_th k_ph")

        x = r * sp.sin(th) * sp.cos(ph)
        y = r * sp.sin(th) * sp.sin(ph)
        z = r * sp.cos(th)

        M_sph_to_xyz = sp.Matrix([[1, 0, 0, 0],\
                                  [0, x.diff(r), x.diff(th), x.diff(ph)],\
                                  [0, y.diff(r), y.diff(th), y.diff(ph)],\
                                  [0, z.diff(r), z.diff(th), z.diff(ph)],\
                                 ])

        v = sp.Matrix([t, r, th , ph])
        k = sp.Matrix([k_t, k_r, k_th, k_ph])
        k_xyz = M_sph_to_xyz*k
        self.convert_k_sph_to_k_xyz = sp.lambdify([t, r, th, ph, k_t, k_r, k_th, k_ph], \
                                     k_xyz, "numpy")

    ##################################################################
    # SPHERICAL - XYZ
    ##################################################################
    def convert_sph4_to_xyz4(self, k_sph, x_sph):

        # t, r, th, ph = x_sph

        # if th < 0.0: 
        #     th = abs(th)
        #     ph += np.pi

        #     x_sph = np.array([t, r, th, ph])
        #     print("DOING", x_sph)

        x_xyz = self.coord_conversion_sph4_to_xyz4(*x_sph)
        k_xyz = self.convert_k_sph_to_k_xyz(*x_sph, *k_sph)
        k_xyz = k_xyz.reshape(*k_sph.shape)
        return k_xyz, x_xyz

    def coord_conversion_sph4_to_xyz4(self, t, r, th, ph):
        z = r*np.cos(th)
        x = r*np.sin(th)*np.cos(ph)
        y = r*np.sin(th)*np.sin(ph)
        return np.array([t, x, y, z])

    ##################################################################
    # XYZ - SPHERICAL
    ##################################################################
    def convert_xyz4_to_sph4(self, k_xyz, x_xyz):
        v_sph = self.coord_conversion_xyz4_to_sph4(*x_xyz)
        k_sph = self.convert_k_xyz_to_k_sph(*x_xyz, *k_xyz)
        k_sph = k_sph.reshape(*k_xyz.shape)
        return k_sph, v_sph

    def coord_conversion_xyz4_to_sph4(self, t, x, y, z):
        r = np.sqrt(x**2 + y**2 + z**2)
        th = np.acos(z/r)
        ph = np.atan2(y, x) #ph = np.atan(y/x)
        return np.array([t, r, th, ph])







import numpy as np
import curvedpy as cp
from curvedpy.geodesics.blackhole import BlackholeGeodesicIntegrator
from curvedpy.utils.utils import getImpactParam
import random
import os
import pickle 
from time import time
from multiprocess import Process, Manager, Pool, cpu_count, current_process #Queue
from functools import partial

from scipy.spatial.transform import Rotation
 
class RelativisticCamera:
    """
    A class to calculate geodesics that can be used for a camera in for example Blender.
    
    ...

    Attributes
    ----------
    gi : BlackholeGeodesicIntegrator
        Geodesic integrator used to calculate geodesics.

    ray_blackhole_hit : np.array
        An array of shape (height,width,samples) that records if the geodesic ends inside the black hole horizon

    geodesics : list
        A list of length height*width*samples containing all geodesics. Each element is an (k_xyz, x_xyz). Here 
        k_xyz is an np.array of 3-momenta (per mass) in cartesian coordinates of the geodesic. And x_xyz is an np.array
        of 3-positions in cartesian coordinates also of the geodesic.

    pixel_coordinates : numpy.array
        An array of shape (height,width,samples). Each element contains the camera pixel coordinates and the sample 
        number (y, x, s = pixel_coordinates[i]). This links the camera coordinates and sample to the list of geodesics.
        For example y, x, s = pixel_coordinates[i] gives the pixel height and width coordinates (y and x) and those belong
        to the geodesic k_xyz, x_xyz = geodesics[i].

    results : list
        A list of length height*width*samples containing the full results of all geodesic integrations

    Methods
    -------
    run_mp(verbose=False, cores = cpu_count(), *args, **kargs)
        Calculates geodesics for all photons leaving the camera.
    
    filename_suggestion()
        Returns a string using all properties of the camera which can be used for saving.

    get_camera_vis_properties(scale=5)
        Returns camera_location, corner_points and corner_rays that can be used to plot the orientation of the camera.
        # PLOTTING EXAMPLE!
        # ax = plt.figure().add_subplot(projection='3d')
        # ax.view_init(elev=0, azim=90, roll=0)
        # lim = 5
        # ax.set_xlabel("x")
        # ax.set_xlim(-lim,lim)
        # ax.set_ylim(-lim,lim)
        # ax.set_zlim(20,31)

        # for c in corner_rays:
        #     ax.quiver(*camera_location, *c, color="k")
        # ax.plot(*list(zip(*corner_points)))

    """


    start_R = Rotation.from_euler('x', 0, degrees=True)

    def __init__(self,  camera_location = np.array([0.0001, 0, 30]), \
                        camera_rotation_euler_props=['x', 0], \
                        resolution = [64, 64],\
                        field_of_view = [0.3, 0.3],\
                        coordinates = "SPH2PATCH",\
                        M=1.0, \
                        a = 0.0,\
                        theta_switch = 0.1*np.pi,\
                        samples = 1,\
                        sampling_seed = 43,\
                        y_lim = [], x_lim = [],\
                        max_step = 1.0,\
                        force_no_sampling = False,\
                        loadFromFile = "",\
                        verbose=True,\
                        verbose_integrator = False,\
                        verbose_init = True):
        """Create a camera object for calculation of geodesics starting in the camera.

            Keyword arguments:
            camera_location -- camera location (default np.array([0.0001, 0, 30]) )
            camera_rotation_euler_props -- Euler rotation properties giving the orientation of the camera (default ['x', 0])
            resolution -- (default [64, 64])
            field_of_view -- (default [0.3, 0.3])
            coordinates -- (default "SPH2PATCH")
            M -- (default 1.0)
            a -- (default 0.0)
            theta_switch -- (default 0.1*np.pi)
            samples -- (default 1)
            sampling_seed -- (default 43)
            y_lim -- (default [])
            x_lim -- (default [])
            max_step -- (default 1.0)
            force_no_sampling -- (default False)
            loadFromFile --
            verbose -- (default True)
            verbose_integrator -- (default False)
            verbose_init -- (default True)
        """

        self.verbose = verbose
        
        if os.path.isfile(loadFromFile):
            self.load(loadFromFile)
        else:

            self.M = M
            self.a = a
            self.theta_switch = theta_switch

            self.camera_location = camera_location
            self.camera_rotation_euler_props = camera_rotation_euler_props
            self.camera_rotation_euler = Rotation.from_euler(camera_rotation_euler_props[0], camera_rotation_euler_props[1], degrees=True)
            self.camera_rotation_matrix = self.camera_rotation_euler.as_matrix()

            self.field_of_view = field_of_view
            self.field_of_view_x, self.field_of_view_y = self.field_of_view

            self.height, self.width = resolution

            if len(y_lim) == 0: self.y_lim = [0, self.height]
            else: self.y_lim = y_lim
            if len(x_lim) == 0: self.x_lim = [0, self.width]
            else: self.x_lim = x_lim

            self.aspectratio = self.height/self.width
            
            self.dy = self.aspectratio/self.height  
            self.dx = 1/self.width  

            self.force_no_sampling = force_no_sampling
            random.seed(sampling_seed)  
            self.sampling_seed = sampling_seed
            self.samples = samples
            self.N = self.samples*self.width*self.height

            self.max_step = max_step
            if coordinates == "":
                coordinates = "SPH2PATCH"
            self.coordinates = coordinates
        


        # if self.theta_switch == "":
        #     self.gi = BlackholeGeodesicIntegrator(mass=self.M, a=self.a, coordinates=self.coordinates, verbose = verbose_integrator)

        # else:
        self.gi = BlackholeGeodesicIntegrator(mass=self.M, a=self.a, theta_switch = self.theta_switch, coordinates=self.coordinates, verbose = verbose_integrator)

        self.results = None

        if verbose_init:
            print("Camera Settings: ")
            print(f"  - {self.coordinates=}")
            print(f"  - {self.M=}")
            print(f"  - {self.a=}")
            print(f"  - {self.theta_switch=}")
            print(f"  - {self.verbose=}")
            print(f"  - {verbose_init=}")
            print(f"  - {self.camera_location=}")
            print(f"  - {self.camera_rotation_euler_props=}")
            print(f"  - {self.width=}")
            print(f"  - {self.height=}")
            print(f"  - {field_of_view=}")
            print(f"  - {self.force_no_sampling=}")
            print(f"  - {self.samples=}")
            print(f"  - {self.sampling_seed=}")
            print(f"  - {self.y_lim=}")
            print(f"  - {self.x_lim=}")
            print(f"  - {self.max_step=}")
            #print(f"  - {force_schwarzschild_integrator=}")
            print("--")



    def filename_suggestion(self):
        """Constructs and returns a sting containing information of the camera."""

        fn = "coordinates_"+str(self.coordinates)+"_res_"+str(self.height)+"x"+str(self.width)+\
                "_fov-x_"+str(self.field_of_view_x)+"_fov-y_"+str(self.field_of_view_y)+\
                "_sample_"+str(self.samples)+"_sampling_seed_"+str(self.sampling_seed)+\
                "_a_"+str(self.a)+"_M_"+str(self.M)+\
                "_xyz0_"+str(self.camera_location[0])+"_"+str(self.camera_location[1])+"_"+str(self.camera_location[2])+\
                f"_rot_{self.camera_rotation_euler_props[0]}_angle_{self.camera_rotation_euler_props[1]}"+\
                "_max_step_"+str(self.max_step)+"_th_switch_"+str(round(self.theta_switch,3))

        return fn

    def cam_information_dict(self):
        """Internal use"""

        return \
            {"version_curvedpy": cp.__version__, "M":self.M, "a":self.a, "camera_location": self.camera_location, \
            "camera_rotation_euler_props":self.camera_rotation_euler_props, \
            "field_of_view": self.field_of_view, \
            "height":self.height, "width":self.width,\
            "y_lim":self.y_lim, "x_lim":self.x_lim,\
            "samples":self.samples, "sampling_seed":self.sampling_seed, \
            "max_step":self.max_step,"coordinates":self.coordinates,\
            "theta_switch":self.theta_switch, "force_no_sampling":self.force_no_sampling\
            }
    
    def store_info_dict(self, info):
        """Dont use this"""

        #print(info)

        if cp.__version__ != info["version_curvedpy"]:
            print("WARNING DIFFERENT CURVEDPY VERSIONS!")

        self.M = info["M"]
        self.a = info["a"]
        self.camera_location = info["camera_location"]
        self.camera_rotation_euler_props = info["camera_rotation_euler_props"]
        self.field_of_view = info["field_of_view"]
        self.height = info["height"]
        self.width = info["width"]
        self.y_lim = info["y_lim"]
        self.x_lim = info["x_lim"]
        self.samples = info["samples"]
        self.sampling_seed = info["sampling_seed"]
        self.max_step = info["max_step"]
        self.coordinates = info["coordinates"]
        if "theta_switch" in info.keys():
            self.theta_switch = info["theta_switch"]
        else:
            self.theta_switch = 0.1*np.pi
        if "force_no_sampling" in info.keys():
            self.force_no_sampling = info["force_no_sampling"]
        else:
            self.force_no_sampling = False

    def get_x_render(self, x):
        """Dont use this"""

        return self.field_of_view_x * (x-int(self.width/2))/self.width

    def get_y_render(self, y):
        """Dont use this"""

        return self.field_of_view_y * (y-int(self.height/2))/self.height * self.aspectratio 

    def get_ray_direction(self, y_render, x_render):
        """Dont use this"""

        #ray_direction = np.array( [ x_render, y_render, -1 ] )
        if self.force_no_sampling:
            ranX, ranY = 0.5, 0.5
        else:
            ranX, ranY = random.random(), random.random()

        ray_direction = np.array( [ x_render + self.dx*(ranX-0.5), y_render + self.dy*(ranY-0.5), -1 ] )
        # The ray direction relative to the camera
        ray_direction = self.camera_rotation_matrix @ ray_direction
        # Normalize the direction ray
        ray_direction = ray_direction / np.linalg.norm(ray_direction)
        return ray_direction

    def get_start_values(self, verbose=False, verbose_lvl = 1):
        """Creates starting values for the geodesics making up the camera."""

        k0, x0, pixel_coordinates = [], [], []

        for s in range(self.samples):
            for y in range(self.height):
                if y >= self.y_lim[0] and y <= self.y_lim[1]:
                    y_render = self.get_y_render(y)#self.field_of_view_y * (y-int(self.height/2))/self.height * self.aspectratio 
                    for x  in range(self.width):
                        if x >= self.x_lim[0] and x <= self.x_lim[1]:
                            x_render = self.get_x_render(x)#self.field_of_view_x * (x-int(self.width/2))/self.width

                            # The ray direction in the -z direction:
                            # ray_direction = np.array( [ x_render + self.dx*(random.random()-0.5), y_render + self.dy*(random.random()-0.5), -1 ] )
                            # ray_direction = np.array( [ x_render, y_render, -1 ] )

                            # # The ray direction relative to the camera
                            # ray_direction = self.camera_rotation_matrix @ ray_direction
                            # # Normalize the direction ray
                            # ray_direction = ray_direction / np.linalg.norm(ray_direction)

                            k0.append(self.get_ray_direction(y_render, x_render))
                            x0.append(self.camera_location)
                            pixel_coordinates.append((y,x,s))

        return k0, x0, pixel_coordinates

    def get_camera_vis_properties(self, scale=5):
        """Returns information to plot the camera orientation."""

        corners = [(0,0), (0,self.width), (self.height, self.width), (self.height, 0)]
        corner_rays = [scale*self.get_ray_direction(self.get_y_render(y), self.get_x_render(x)) for y, x in corners]
        corner_points = [self.camera_location+c for c in [*corner_rays, corner_rays[0]]] # Add the first one to make a closed loop for plotting

        return self.camera_location, corner_points, corner_rays


    def run_mp(self, verbose=False, cores = cpu_count(), *args, **kargs):
        """Calculate all geodesics starting in the camera."""

        start_time = time()

        self.ray_blackhole_hit = np.zeros(self.width * self.height * self.samples * 1)
        self.ray_blackhole_hit.shape = self.height, self.width, self.samples

        # Geodesics have different lengths, so we use lists to save them
        self.geodesics = []

        if self.verbose: print(f"Starting run_mp on {cores=}")
        k0, x0, self.pixel_coordinates = self.get_start_values()
        if self.verbose: print(f"  We have {len(k0)} models to run at {self.height}x{self.width} with sampling {self.samples}")

        split_factor = 16
        if len(k0) < split_factor:
            split_factor = 1
        start_values = list(zip(np.array_split(k0, cores*split_factor), np.array_split(x0, cores*split_factor)))
        
        if self.verbose: print(f"Multiproc info: {cores=}, {split_factor=}, {len(start_values)}, {len(start_values[0])}")
        if self.verbose: print()

        def wrap_calc_trajectory(k0_xyz, x0_xyz, shared, mes="no mes"):
            #if self.verbose: print(f"          Starting: {current_process().name}, processing array of shape: {k0_xyz.shape}")
            res = shared['gi'].geodesic(k0_xyz = k0_xyz, x0_xyz = x0_xyz, max_step = self.max_step, full_save=False, verbose=False, *args, **kargs)
            #if self.verbose: print(f"          Done: {current_process()}")
            return res

        with Manager() as manager:
            shared = manager.dict()
            shared['gi'] = self.gi
            partial_wrap_calc_trajectory = partial(wrap_calc_trajectory, shared=shared)
            with Pool(cores) as pool:
                results_pool = pool.starmap(partial_wrap_calc_trajectory, start_values)

        self.results = [ x for xs in results_pool for x in xs]

        for i in range(len(self.pixel_coordinates)):
            y, x, s = self.pixel_coordinates[i]
            
            if isinstance(self.results[i][2], list):
                self.ray_blackhole_hit[y, x, s] = int(self.results[i][2][-1].hit_blackhole)
            else:
                self.ray_blackhole_hit[y, x, s] = int(self.results[i][2]["hit_blackhole"])
            k_xyz, x_xyz = self.results[i][0], self.results[i][1]
            self.geodesics.append([k_xyz, x_xyz])

        if self.verbose: print(f"  Running is done and we have {len(self.results)} results")
        if self.verbose: print(f"Time taken for run_mp: {time()-start_time}")


    def save(self, fname, directory):
        """Saves the camera outputs and info into a pickle file."""

        if os.path.isdir(directory):
            # with open(os.path.join(directory, fname+'.pkl'), 'wb') as f:
            #     info = self.cam_information_dict()
            #     pickle.dump({"info": info, "geodesics": self.geodesics, "pixel_coordinates": self.pixel_coordinates, \
            #         "results": self.results, "ray_blackhole_hit": self.ray_blackhole_hit}, f)

            with open(os.path.join(directory, fname+'.pkl'), 'wb') as f:
                info = self.cam_information_dict()
                pickle.dump({"info": info, "geodesics": self.geodesics, "pixel_coordinates": self.pixel_coordinates, \
                    "ray_blackhole_hit": self.ray_blackhole_hit}, f)

        else:
            print("dir not found")

    def load(self, filepath):
        """Loads camera results and information into this object."""

        if os.path.isfile(filepath):
            with open(filepath, 'rb') as f:
                filecontent = pickle.load(f)
                print(filecontent.keys())
                self.store_info_dict(filecontent["info"])
                self.geodesics = filecontent["geodesics"]
                self.ray_blackhole_hit = filecontent["ray_blackhole_hit"]
                self.pixel_coordinates = filecontent["pixel_coordinates"]
        else:

            print(f"CAM: FILE NOT FOUND {filepath}")




    def calcStats(self, select_by_impact_parameter = [0, np.inf]):
        """Debug function"""

        stats = []
        for el in self.results:
            k, x, res = el
            x0 = np.column_stack(x)[0]
            #print(x0)
            k0 = np.column_stack(k)[0]
            impact_vector_normed, impact_par = getImpactParam(x0, k0)
            if impact_par >= select_by_impact_parameter[0] and impact_par <= select_by_impact_parameter[1]:
                x_end = np.column_stack(x)[-1]
                k_end = np.column_stack(k)[-1]
                deflect = k_end.dot(k0)
                hit = int(res.hit_blackhole)
                stats.append((x0, k0, x_end, k_end, impact_vector_normed, impact_par, deflect, hit))

        x0, k0, x_end, k_end, impact_vector_normed, impact_par, deflect, hit = list(zip(*stats))
        
        return {"x0":x0, "k0":k0, "x_end":x_end, "k_end":k_end, "impact_vector_normed":impact_vector_normed, \
                "impact_par":impact_par, "deflect":deflect, "hit":hit}


# https://en.wikipedia.org/wiki/Euler_angles
# https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.transform.Rotation.from_euler.html#scipy.spatial.transform.Rotation.from_euler
#start_R = Rotation.from_euler(seq='XYZ', angles=[0, 90, 0], degrees=True)


# Or do the follwoing:
# np.array(cameuler.to_matrix())
# np.array(bpy.data.scenes['Scene'].camera.matrix_world.to_euler().to_matrix())@np.array([0,0,-1])

# The standard direction from which the rotation is measured in Blender is the 0,0,-1 direction, So downwards in the z direction


# Do the following inside blender:
# Get the Euler rotation of the camera:
# cam_euler = bpy.data.scenes['Scene'].camera.matrix_world.to_euler()
# Put this rotation in a scipy rotation:
# r = Rotation.from_euler(cam_euler.order, [cam_euler.x, cam_euler.y, cam_euler.z], degrees=False)
# Give this r as camera_rotation to this class


# For more information:
# Eurler rotations in Blender are given like this:
# cameuler = C.scene.camera.matrix_world.to_euler()
# > Euler((1.1093189716339111, -0.0, 0.8149281740188599), 'XYZ')
# (https://blender.stackexchange.com/questions/130948/blender-api-get-current-location-and-rotation-of-camera-tracking-an-object)
# cameuler.order gives the 'XYZ'
# 

# In scipy a rotation can be created using:
# r = Rotation.from_euler('zyx', [90, 45, 30], degrees=True)



import numpy as np
from scipy.interpolate import splev, splprep



# splprep: 
# https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.splprep.html#scipy.interpolate.splprep
# splev: 
# https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.splev.html#scipy.interpolate.splev
class Curve:
	def __init__(self, t, x, y, z, k=3, verbose=False):
		self.tck, u = splprep([x, y, z], u=t, k=5)

	def get(self, t):
		x, y, z = splev(t, self.tck)
		return x, y, z



import numpy as np
from curvedpy.geodesics.blackhole import BlackholeGeodesicIntegrator
from curvedpy.utils.utils import getImpactParam
import random
import os
import pickle 
import time
from multiprocess import Process, Manager, Pool, cpu_count #Queue
from functools import partial


#import mathutils # I do not want to use this in the end but need to check with how Blender rotates things compared to scipy
from scipy.spatial.transform import Rotation
 
class RelativisticCamera:

    # https://en.wikipedia.org/wiki/Euler_angles
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.transform.Rotation.from_euler.html#scipy.spatial.transform.Rotation.from_euler
    #start_R = Rotation.from_euler(seq='XYZ', angles=[0, 90, 0], degrees=True)
    start_R = Rotation.from_euler('x', 0, degrees=True)
    #rot = Rotation.from_euler('x', 45, degrees=True)#.as_matrix()

    def __init__(self,  camera_location = np.array([0.0001, 0, 30]), \
                        camera_rotation_euler_props=['x', 0], \
                        resolution = [64, 64],\
                        field_of_view = [0.3, 0.3],\
                        coordinates = "",\
                        M=1.0, \
                        a = 0.0,\
                        theta_switch = "",\
                        samples = 1,\
                        sampling_seed = 43,\
                        y_lim = [], x_lim = [],\
                        max_step = np.inf,\
                        verbose=False,\
                        verbose_init = True):

        self.verbose = verbose

        self.M = M
        self.a = a
        self.theta_switch = theta_switch

        self.camera_location = camera_location
        self.camera_rotation_euler_props = camera_rotation_euler_props
        self.camera_rotation_euler = Rotation.from_euler(camera_rotation_euler_props[0], camera_rotation_euler_props[1], degrees=True)
        self.camera_rotation_matrix = self.camera_rotation_euler.as_matrix()

        self.field_of_view = field_of_view
        self.field_of_view_x, self.field_of_view_y = self.field_of_view

        self.height, self.width = resolution

        if len(y_lim) == 0: self.y_lim = [0, self.height]
        else: self.y_lim = y_lim
        if len(x_lim) == 0: self.x_lim = [0, self.width]
        else: self.x_lim = x_lim

        self.aspectratio = self.height/self.width
        
        self.dy = self.aspectratio/self.height  
        self.dx = 1/self.width  
        random.seed(sampling_seed)  
        self.samples = samples
        self.N = self.samples*self.width*self.height

        self.max_step = max_step
        if coordinates == "":
            coordinates = "SPH2PATCH"
        self.coordinates = coordinates


        if self.theta_switch == "":
            self.gi = BlackholeGeodesicIntegrator(mass=self.M, a=self.a, coordinates=self.coordinates, verbose = self.verbose)
        else:
            self.gi = BlackholeGeodesicIntegrator(mass=self.M, a=self.a, theta_switch = self.theta_switch, coordinates=self.coordinates, verbose = self.verbose)

        self.results = None

        if verbose_init:
            print("Camera Settings: ")
            print(f"  - {self.coordinates=}")
            print(f"  - {self.M=}")
            print(f"  - {self.a=}")
            print(f"  - {self.theta_switch=}")
            print(f"  - {self.verbose=}")
            print(f"  - {verbose_init=}")
            print(f"  - {self.camera_location=}")
            print(f"  - {self.camera_rotation_euler_props=}")
            print(f"  - {resolution=}")
            print(f"  - {field_of_view=}")
            print(f"  - {self.samples=}")
            print(f"  - {sampling_seed=} not supported")
            print(f"  - {self.y_lim=}")
            print(f"  - {self.x_lim=}")
            print(f"  - {self.max_step=}")
            #print(f"  - {force_schwarzschild_integrator=}")
            print("--")



    def filename_suggestion(self):
        fn = "coordinates_"+str(self.coordinates)+"_res_"+str(self.height)+"x"+str(self.width)+\
                "_fov-x_"+str(self.field_of_view_x)+"_fov-y_"+str(self.field_of_view_y)+\
                "_a_"+str(self.a)+"_M_"+str(self.M)+\
                "_xyz0_"+str(self.camera_location[0])+"_"+str(self.camera_location[1])+"_"+str(self.camera_location[2])+\
                f"_rot_{self.camera_rotation_euler_props[0]}_angle_{self.camera_rotation_euler_props[1]}"+\
                "_max_step_"+str(self.max_step)
                #"_rot_"+rot_axis+"_angle_"+str(degrees)

        # if self.schwarzschild_integrator:
        #     fn += "_forceSS_"+str(self.schwarzschild_integrator)
        return fn

    def cam_information_dict(self):
        return \
            {"M":self.M, "a":self.a, "camera_location": self.camera_location, \
            "camera_rotation_euler_props":self.camera_rotation_euler_props, \
            "field_of_view": self.field_of_view, \
            "height":self.height, "width":self.width,\
            "y_lim":self.y_lim, "x_lim":self.x_lim,\
            "max_step":self.max_step,"coordinates":self.coordinates,\
            }

    def get_x_render(self, x):
        return self.field_of_view_x * (x-int(self.width/2))/self.width
    def get_y_render(self, y):
        return self.field_of_view_y * (y-int(self.height/2))/self.height * self.aspectratio 
    def get_ray_direction(self, y_render, x_render):
        ray_direction = np.array( [ x_render, y_render, -1 ] )
        # The ray direction relative to the camera
        ray_direction = self.camera_rotation_matrix @ ray_direction
        # Normalize the direction ray
        ray_direction = ray_direction / np.linalg.norm(ray_direction)
        return ray_direction

    def get_start_values(self, verbose=False, verbose_lvl = 1):

        self.ray_end = np.zeros(self.width * self.height * 6) # The six is for 3 end coordinates and 3 end directions
        self.ray_end.shape = self.height, self.width, 6

        self.ray_blackhole_hit = np.zeros(self.width * self.height * 1)
        self.ray_blackhole_hit.shape = self.height, self.width


        k0, x0, pixel_coordinates = [], [], []

        for s in range(self.samples):
            for y in range(self.height):
                if y >= self.y_lim[0] and y <= self.y_lim[1]:
                    y_render = self.get_y_render(y)#self.field_of_view_y * (y-int(self.height/2))/self.height * self.aspectratio 
                    for x  in range(self.width):
                        if x >= self.x_lim[0] and x <= self.x_lim[1]:
                            x_render = self.get_x_render(x)#self.field_of_view_x * (x-int(self.width/2))/self.width

                            # The ray direction in the -z direction:
                            # ray_direction = np.array( [ x_render + self.dx*(random.random()-0.5), y_render + self.dy*(random.random()-0.5), -1 ] )
                            # ray_direction = np.array( [ x_render, y_render, -1 ] )

                            # # The ray direction relative to the camera
                            # ray_direction = self.camera_rotation_matrix @ ray_direction
                            # # Normalize the direction ray
                            # ray_direction = ray_direction / np.linalg.norm(ray_direction)

                            k0.append(self.get_ray_direction(y_render, x_render))
                            x0.append(self.camera_location)
                            pixel_coordinates.append((y,x))

        return k0, x0, pixel_coordinates

    #def get_start_values_radial_line(self, verbose=False):

    def get_camera_vis_properties(self, scale=5):
        corners = [(0,0), (0,self.width), (self.height, self.width), (self.height, 0)]
        corner_rays = [scale*self.get_ray_direction(self.get_y_render(y), self.get_x_render(x)) for y, x in corners]
        corner_points = [self.camera_location+c for c in [*corner_rays, corner_rays[0]]] # Add the first one to make a closed loop for plotting

        # PLOTTING EXAMPLE!
        # ax = plt.figure().add_subplot(projection='3d')
        # ax.view_init(elev=0, azim=90, roll=0)
        # lim = 5
        # ax.set_xlabel("x")
        # ax.set_xlim(-lim,lim)
        # ax.set_ylim(-lim,lim)
        # ax.set_zlim(20,31)

        # for c in corner_rays:
        #     ax.quiver(*camera_location, *c, color="k")
        # ax.plot(*list(zip(*corner_points)))

        return self.camera_location, corner_points, corner_rays


    def run_mp(self, verbose=False, cores = cpu_count(), *args, **kargs):
        self.ray_end = np.zeros(self.width * self.height * 6) # The six is for 3 end coordinates and 3 end directions
        self.ray_end.shape = self.height, self.width, 6

        self.ray_blackhole_hit = np.zeros(self.width * self.height * 1)
        self.ray_blackhole_hit.shape = self.height, self.width

        # Geodesics have different lengths, so we use lists to save them
        self.geodesics = []

        if verbose: print(f"Starting run_mp on {cores=}")
        k0, x0, self.pixel_coordinates = self.get_start_values()
        if verbose: print(f"  We have {len(k0)} models to run at {self.height}x{self.width}")

        start_values = list(zip(np.array_split(k0, cores), np.array_split(x0, cores)))
        
        def wrap_calc_trajectory(k0_xyz, x0_xyz, shared, mes="no mes"):
            res = shared['gi'].geodesic(k0_xyz = k0_xyz, x0_xyz = x0_xyz, max_step = self.max_step, verbose=False, *args, **kargs)
            return res

        with Manager() as manager:
            shared = manager.dict()
            shared['gi'] = self.gi
            #_ = time.time()
            partial_wrap_calc_trajectory = partial(wrap_calc_trajectory, shared=shared)
            with Pool(cores) as pool:
                results_pool = pool.starmap(partial_wrap_calc_trajectory, start_values)

        self.results = [ x for xs in results_pool for x in xs]
        #self.ray_end[y, x, 0:3] = list(zip(*x_xyz))[-1]
        #self.ray_end[y, x, 3:6] = list(zip(*k_xyz))[-1]

        for i in range(len(self.pixel_coordinates)):
            y, x = self.pixel_coordinates[i]
            # if y==16 and x == 32:
            #     print(y, x, self.results[i][2])#["hit_blackhole"], )

            if isinstance(self.results[i][2], list):
                self.ray_blackhole_hit[y, x] = int(self.results[i][2][-1].hit_blackhole)
            else:
                self.ray_blackhole_hit[y, x] = int(self.results[i][2]["hit_blackhole"])
            k_xyz, x_xyz = self.results[i][0], self.results[i][1]
            self.geodesics.append([k_xyz, x_xyz])
            #print("plplpl", x_xyz)

            #print(f"Strating point is {list(zip(*x_xyz))[0]} and should be {x0[i]}")
            #print(f"Strating k is {list(zip(*k_xyz))[0]} and should be {k0[i]}")

            self.ray_end[y, x, 0:3] = list(zip(*x_xyz))[-1]
            self.ray_end[y, x, 3:6] = list(zip(*k_xyz))[-1]
            
        #MOET NOG DE RAY_END EN DERGELIJKE SETTEN!
        if verbose: print(f"  Running is done and we have {len(self.results)} results")


    # def run(self, verbose=False, verbose_lvl = 1):

    #     self.ray_end = np.zeros(self.width * self.height * 6) # The six is for 3 end coordinates and 3 end directions
    #     self.ray_end.shape = self.height, self.width, 6

    #     self.ray_blackhole_hit = np.zeros(self.width * self.height * 1)
    #     self.ray_blackhole_hit.shape = self.height, self.width



    #     #camera_locations = []
    #     #ray_directions = []

    #     self.results = []
    #     start = time.time()
    #     for s in range(self.samples):
    #         for y in range(self.height):
    #             if y%verbose_lvl == 0 and verbose:
    #                 print(" Status, starting y: ", y, ", elap time: ", round(time.time()-start, 5))
    #             y_render = self.field_of_view_y * (y-int(self.height/2))/self.height * self.aspectratio 
                
    #             # temp. store the geodesics in the width direction

    #             for x  in range(self.width):

    #                 #camera_locations.append(self.camera_location)

    #                 x_render = self.field_of_view_x * (x-int(self.width/2))/self.width

    #                 # The ray direction in the -z direction:
    #                 # ray_direction = np.array( [ x_render + self.dx*(random.random()-0.5), y_render + self.dy*(random.random()-0.5), -1 ] )
    #                 ray_direction = np.array( [ x_render, y_render, -1 ] )

    #                 # The ray direction relative to the camera
    #                 ray_direction = self.camera_rotation_matrix @ ray_direction
    #                 # Normalize the direction ray
    #                 ray_direction = ray_direction / np.linalg.norm(ray_direction)

    #                 #ray_directions.append(ray_direction)

    #                 k_xyz, x_xyz, res = self.gi.geodesic( k0_xyz = ray_direction, \
    #                                             x0_xyz = self.camera_location,\
    #                                             R_end = -1,\
    #                                             curve_start = 0, \
    #                                             curve_end = 50, \
    #                                             nr_points_curve = 50, \
    #                                             method = "RK45",\
    #                                             max_step = self.max_step,\
    #                                             first_step = None,\
    #                                             rtol = 1e-3,\
    #                                             atol = 1e-6,\
    #                                             verbose = self.verbose \
    #                                             )
    #                 #print(list(zip(*x_xyz))[0], list(zip(*k_xyz))[0])
    #                 #print(x_xyz.shape, x_xyz[-1].shape)

    #                 self.ray_end[y, x, 0:3] = list(zip(*x_xyz))[-1]# x_xyz[-1]#, *k_xyz[-1]]
    #                 self.ray_end[y, x, 3:6] = list(zip(*k_xyz))[-1]

    #                 self.ray_blackhole_hit[y, x] = int(res.hit_blackhole)

    #                 self.results.append([k_xyz, x_xyz, res])
    #                 geodesics_width.append([k_xyz, x_xyz])

    #             # Save the goedesics in the width direction into the global list of geodesics
    #             self.geodesics.append(geodesics_width)

    def save(self, fname, directory):
        if os.path.isdir(directory):
            with open(os.path.join(directory, fname+'.pkl'), 'wb') as f:
                info = self.cam_information_dict()
                pickle.dump({"info": info, "geodesics": self.geodesics, "pixel_coordinates": self.pixel_coordinates, \
                    "results": self.results, "ray_end": self.ray_end, "ray_blackhole_hit": self.ray_blackhole_hit}, f)

        else:
            print("dir not found")

    def save_minimal(self, fname, directory="."):
        extension = ".csv"
        if os.path.isdir(directory):
            x, y, z = self.ray_end[:,:,0], self.ray_end[:,:,1], self.ray_end[:,:,2]
            kx, ky, kz = self.ray_end[:,:,3], self.ray_end[:,:,4], self.ray_end[:,:,5]

            h = "Cartesian x component in 3-space of end direction of a photon. \n"+\
                "Rows are camera height direction, columns are width direction. \n"+\
                "You can, for example, load this file using numpy.loadtxt. \n"+\
                str(self.cam_information_dict())
            np.savetxt(os.path.join(directory, 'k_x_'+fname+extension), kx, header = h)
            h = "Cartesian y component in 3-space of end direction of a photon. \n"+\
                "Rows are camera height direction, columns are width direction. \n"+\
                "You can, for example, load this file using numpy.loadtxt. \n"+\
                str(self.cam_information_dict())
            np.savetxt(os.path.join(directory, 'k_y_'+fname+extension), ky, header=h)
            h = "Cartesian z component in 3-space of end direction of a photon. \n"+\
                "Rows are camera height direction, columns are width direction. \n"+\
                "You can, for example, load this file using numpy.loadtxt. \n"+\
                str(self.cam_information_dict())
            np.savetxt(os.path.join(directory, 'k_z_'+fname+extension), kz, header=h)

            h = "Cartesian x component in 3-space of end point of a photon. \n"+\
                "Rows are camera height direction, columns are width direction. \n"+\
                "You can, for example, load this file using numpy.loadtxt. \n"+\
                str(self.cam_information_dict())
            np.savetxt(os.path.join(directory, 'x_'+fname+extension), x, header=h)
            h = "Cartesian y component in 3-space of end point of a photon. \n"+\
                "Rows are camera height direction, columns are width direction. \n"+\
                "You can, for example, load this file using numpy.loadtxt. \n"+\
                str(self.cam_information_dict())
            np.savetxt(os.path.join(directory, 'y_'+fname+extension), y, header=h)
            h = "Cartesian z component in 3-space of end point of a photon. \n"+\
                "Rows are camera height direction, columns are width direction. \n"+\
                "You can, for example, load this file using numpy.loadtxt. \n"+\
                str(self.cam_information_dict())
            np.savetxt(os.path.join(directory, 'z_'+fname+extension), z, header=h)

            h = "Hit blackhole: 1 if the photon end in the black hole and 0 if not. \n"+\
                "Rows are camera height direction, columns are width direction. \n"+\
                "You can, for example, load this file using numpy.loadtxt. \n"+\
                str(self.cam_information_dict())
            np.savetxt(os.path.join(directory, 'hit_blackhole'+extension), self.ray_blackhole_hit, header=h)

            with open(os.path.join(directory, "full_geodesics_"+fname+'.pkl'), 'wb') as f:
                k, x, res = list(zip(*self.results))
                pickle.dump({"cam_info": self.cam_information_dict(), "k":k, "x":x }, f)

    


    def load(self, filepath):
        if os.path.isfile(filepath):
            with open(filepath, 'rb') as f:
                filecontent = pickle.load(f)
                print(filecontent.keys())
                self.results = filecontent["results"]
                self.ray_end = filecontent["ray_end"]
                self.ray_blackhole_hit = filecontent["ray_blackhole_hit"]
                self.pixel_coordinates = filecontent["pixel_coordinates"]
        else:

            print(f"CAM: FILE NOT FOUND {filepath}")




    # DEBUG FUNCTIONS
    def calcStats(self, select_by_impact_parameter = [0, np.inf]):
        stats = []
        for el in self.results:
            k, x, res = el
            x0 = np.column_stack(x)[0]
            #print(x0)
            k0 = np.column_stack(k)[0]
            impact_vector_normed, impact_par = getImpactParam(x0, k0)
            if impact_par >= select_by_impact_parameter[0] and impact_par <= select_by_impact_parameter[1]:
                x_end = np.column_stack(x)[-1]
                k_end = np.column_stack(k)[-1]
                deflect = k_end.dot(k0)
                hit = int(res.hit_blackhole)
                stats.append((x0, k0, x_end, k_end, impact_vector_normed, impact_par, deflect, hit))

        x0, k0, x_end, k_end, impact_vector_normed, impact_par, deflect, hit = list(zip(*stats))
        
        return {"x0":x0, "k0":k0, "x_end":x_end, "k_end":k_end, "impact_vector_normed":impact_vector_normed, \
                "impact_par":impact_par, "deflect":deflect, "hit":hit}

    #def deflectImage(self):



# Or do the follwoing:
# np.array(cameuler.to_matrix())
# np.array(bpy.data.scenes['Scene'].camera.matrix_world.to_euler().to_matrix())@np.array([0,0,-1])

# The standard direction from which the rotation is measured in Blender is the 0,0,-1 direction, So downwards in the z direction


# Do the following inside blender:
# Get the Euler rotation of the camera:
# cam_euler = bpy.data.scenes['Scene'].camera.matrix_world.to_euler()
# Put this rotation in a scipy rotation:
# r = Rotation.from_euler(cam_euler.order, [cam_euler.x, cam_euler.y, cam_euler.z], degrees=False)
# Give this r as camera_rotation to this class


# For more information:
# Eurler rotations in Blender are given like this:
# cameuler = C.scene.camera.matrix_world.to_euler()
# > Euler((1.1093189716339111, -0.0, 0.8149281740188599), 'XYZ')
# (https://blender.stackexchange.com/questions/130948/blender-api-get-current-location-and-rotation-of-camera-tracking-an-object)
# cameuler.order gives the 'XYZ'
# 

# In scipy a rotation can be created using:
# r = Rotation.from_euler('zyx', [90, 45, 30], degrees=True)
